"""Get Netscaler details."""
import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer import utils


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Get Netscaler details."""
    mgr = SoftLayer.LoadBalancerManager(env.client)

    this_lb = mgr.get_adc(identifier)
    table = netscaler_table(this_lb)
    env.fout(table)


def netscaler_table(this_lb):
    """Formats the netscaler info table"""
    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'
    table.add_row(['Id', this_lb.get('id')])
    table.add_row(['Type', this_lb.get('description')])
    table.add_row(['Name', this_lb.get('name')])
    table.add_row(['Location', utils.lookup(this_lb, 'datacenter', 'longName')])
    table.add_row(['Managment Ip', this_lb.get('managementIpAddress')])
    table.add_row(['Root Password', utils.lookup(this_lb, 'password', 'password')])
    table.add_row(['Primary Ip', this_lb.get('primaryIpAddress')])
    table.add_row(['License Expiration', utils.clean_time(this_lb.get('licenseExpirationDate'))])
    subnet_table = formatting.Table(['Id', 'Subnet', 'Type', 'Space'])
    for subnet in this_lb.get('subnets', []):
        subnet_table.add_row([
            subnet.get('id'),
            f"{subnet.get('networkIdentifier')}/{subnet.get('cidr')}",
            subnet.get('subnetType'),
            subnet.get('addressSpace')
        ])
    table.add_row(['Subnets', subnet_table])

    vlan_table = formatting.Table(['Id', 'Number'])
    for vlan in this_lb.get('networkVlans', []):
        vlan_table.add_row([vlan.get('id'), vlan.get('vlanNumber')])
    table.add_row(['Vlans', vlan_table])

    return table



getObject = {
    'id': 1234,
    'billingItem': {
        'id': 1056
    },
    'number': 999,
    'networkIdentifier': '1.2.3.4',
    'cidr': '26',
    'subnetType': 'ADDITIONAL_PRIMARY',
    'networkVlan': {
        'networkSpace': 'PUBLIC',
        'primaryRouter': {
            'fullyQualifiedDomainName': 'fcr03a.dal13.softlayer.com'
        }
    },
    'gateway': '1.2.3.254',
    'broadcastAddress': '1.2.3.255',
    'datacenter': {
        'name': 'dal10',
        'id': 1
    },
    'virtualGuests': [
        {
            'hostname': 'hostname0',
            'domain': 'sl.test',
            'primaryIpAddress': '1.2.3.10',
            'primaryBackendIpAddress': '10.0.1.2'
        }
    ],
    'hardware': [],
    'usableIpAddressCount': 22,
    'note': 'test note',
    'tagReferences': [
        {'id': 1000123,
         'resourceTableId': 1234,
         'tag': {'id': 100123,
                 'name': 'subnet: test tag'},
         }
    ],
    'ipAddresses': [
        {'id': 123456,
         'ipAddress': '16.26.26.25'},
        {'id': 123457,
         'ipAddress': '16.26.26.26'}]
}

editNote = True
setTags = True
cancel = True
route = True
clearRoute = True



"""Add or remove specific subnets access for a user."""
# :license: MIT, see LICENSE for more details.


import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.option('--add/--remove', default=True,
              help="Add or remove access to subnets.")
@click.argument('user', nargs=1, required=True)
@click.argument('subnet', nargs=-1, required=True)
@environment.pass_env
def cli(env, user, add, subnet):
    """Add or remove subnets access for a user."""
    mgr = SoftLayer.UserManager(env.client)
    user_id = helpers.resolve_id(mgr.resolve_ids, user, 'username')
    if add:
        result = mgr.vpn_subnet_add(user_id, subnet)
    else:
        result = mgr.vpn_subnet_remove(user_id, subnet)

    if result:
        click.secho(f"{user} updated successfully", fg='green')
    else:
        click.secho(f"Failed to update {user}", fg='red')



"""List hosts with access to file volume."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import columns as column_helper
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers
from SoftLayer.CLI import storage_utils


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('volume_id')
@click.option('--sortby', help='Column to sort by', default='name')
@click.option('--columns',
              callback=column_helper.get_formatter(storage_utils.COLUMNS),
              help=f"Columns to display. Options: {', '.join(column.name for column in storage_utils.COLUMNS)}",
              default=','.join(storage_utils.DEFAULT_COLUMNS))
@environment.pass_env
def cli(env, columns, sortby, volume_id):
    """List ACLs.

    Example::
        slcli file access-list 12345678 --sortby id
        This command lists all hosts that are authorized to access volume with ID 12345678 and sorts them by ID.
"""

    file_manager = SoftLayer.FileStorageManager(env.client)
    resolved_id = helpers.resolve_id(file_manager.resolve_ids, volume_id, 'Volume Id')
    access_list = file_manager.get_file_volume_access_list(
        volume_id=resolved_id)
    table = formatting.Table(columns.columns)
    table.sortby = sortby

    for key, type_name in [('allowedVirtualGuests', 'VIRTUAL'),
                           ('allowedHardware', 'HARDWARE'),
                           ('allowedSubnets', 'SUBNET'),
                           ('allowedIpAddresses', 'IP')]:
        for obj in access_list.get(key, []):
            obj['type'] = type_name
            table.add_row([value or formatting.blank()
                           for value in columns.row(obj)])

    env.fout(table)



"""Get monitoring for a hardware device."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Get details for a hardware monitors device."""

    hardware = SoftLayer.HardwareManager(env.client)

    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'

    monitoring = hardware.get_hardware(identifier)

    table.add_row(['Domain', monitoring.get('fullyQualifiedDomainName')])
    table.add_row(['Public Ip', monitoring.get('primaryIpAddress')])
    table.add_row(['Private Ip', monitoring.get('primaryBackendIpAddress')])
    table.add_row(['Location', monitoring['datacenter']['longName']])

    monitoring_table = formatting.Table(['Id', 'IpAddress', 'Status', 'Type', 'Notify'])
    for monitor in monitoring['networkMonitors']:
        monitoring_table.add_row([monitor.get('id'), monitor.get('ipAddress'), monitor.get('status'),
                                  monitor['queryType']['name'], monitor['responseAction']['actionDescription']])

    table.add_row(['Devices monitors', monitoring_table])

    env.fout(table)



"""Order a block storage replica volume."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import helpers
from SoftLayer import utils


CONTEXT_SETTINGS = {'token_normalize_func': lambda x: x.upper()}


@click.command(cls=SoftLayer.CLI.command.SLCommand, context_settings=CONTEXT_SETTINGS)
@click.argument('volume_id')
@click.option('--datacenter', '-d',
              help='Short name of the datacenter for the replica (e.g.: dal09)',
              required=True)
@click.option('--iops', '-i',
              help='Performance Storage IOPs, between 100 and 6000 in multiples of 100. If no IOPS value is specified,'
              ' the IOPS value of the original volume will be used.',
              type=int)
@click.option('--os-type', '-o',
              help='Operating System Type (eg. LINUX) of the primary volume for '
              'which a replica is ordered [optional].',
              type=click.Choice([
                  'HYPER_V',
                  'LINUX',
                  'VMWARE',
                  'WINDOWS_2008',
                  'WINDOWS_GPT',
                  'WINDOWS',
                  'XEN']))
@click.option('--snapshot-schedule', '-s',
              help='Snapshot schedule to use for replication. Options are: '
              'HOURLY, DAILY, WEEKLY',
              required=True,
              type=click.Choice(['HOURLY', 'DAILY', 'WEEKLY']))
@click.option('--tier', '-t',
              help='Endurance Storage Tier (IOPS per GB) of the primary volume for which a replica is ordered '
              '[optional]. If no tier is specified, the tier of the original volume will be used',
              type=click.Choice(['0.25', '2', '4', '10']))
@environment.pass_env
def cli(env, volume_id, snapshot_schedule, datacenter, tier, os_type, iops):
    """Order a block storage replica volume."""
    block_manager = SoftLayer.BlockStorageManager(env.client)
    block_volume_id = helpers.resolve_id(block_manager.resolve_ids, volume_id, 'Block Volume')

    if tier is not None:
        tier = float(tier)

    if iops is not None:
        if iops < 100 or iops > 6000:
            raise exceptions.ArgumentError(f"Invalid value for '--iops' / '-i': '{iops}' is not one "
                                           "of between 100 and 6000.")
        if iops % 100 != 0:
            raise exceptions.ArgumentError(f"Invalid value for '--iops' / '-i': '{iops}' is not a multiple of 100.")

    try:
        order = block_manager.order_replicant_volume(
            block_volume_id,
            snapshot_schedule=snapshot_schedule,
            location=datacenter,
            tier=tier,
            os_type=os_type,
            iops=iops
        )
    except ValueError as ex:
        raise exceptions.ArgumentError(str(ex))

    if 'placedOrder' in order.keys():
        click.echo(f"Order #{utils.lookup(order, 'placedOrder', 'id')} placed successfully!")
        for item in utils.lookup(order, 'placedOrder', 'items'):
            click.echo(" > %s" % item.get('description'))
    else:
        click.echo("Order could not be placed! Please verify your options " +
                   "and try again.")



"""Manage LBaaS health checks."""
import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.exceptions import SoftLayerAPIError
from SoftLayer import utils


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--uuid', required=True, help="Health check UUID to modify.")
@click.option('--interval', '-i', type=click.IntRange(2, 60), help="Seconds between checks. [2-60]")
@click.option('--retry', '-r', type=click.IntRange(1, 10), help="Number of times before marking as DOWN. [1-10]")
@click.option('--timeout', '-t', type=click.IntRange(1, 59), help="Seconds to wait for a connection. [1-59]")
@click.option('--url', '-u', help="Url path for HTTP/HTTPS checks.")
@environment.pass_env
def cli(env, identifier, uuid, interval, retry, timeout, url):
    """Manage LBaaS health checks.

    Example::

        slcli loadbal health-edit (--lb-id LOADBAL_ID) (--health-uuid HEALTH_CHECK_UUID) [-i, --interval INTERVAL]
        [-r, --retry RETRY] [-t, --timeout TIMEOUT] [-u, --url URL]
    """

    if not any([interval, retry, timeout, url]):
        raise exceptions.ArgumentError("Specify either interval, retry, timeout, url")

    # map parameters to expected API names
    template = {'healthMonitorUuid': uuid, 'interval': interval,
                'maxRetries': retry, 'timeout': timeout, 'urlPath': url}
    # Removes those empty values
    clean_template = {k: v for k, v in template.items() if v is not None}

    mgr = SoftLayer.LoadBalancerManager(env.client)
    # Need to get the LBaaS uuid if it wasn't supplied
    lb_uuid, lb_id = mgr.get_lbaas_uuid_id(identifier)
    print(f"UUID: {lb_uuid}, ID: {lb_id}")

    # Get the current health checks, and find the one we are updating.
    mask = "mask[healthMonitors, listeners[uuid,defaultPool[healthMonitor]]]"
    lbaas = mgr.get_lb(lb_id, mask=mask)

    check = {}
    # Set the default values, because these all need to be set if we are not updating them.
    for listener in lbaas.get('listeners', []):
        if utils.lookup(listener, 'defaultPool', 'healthMonitor', 'uuid') == uuid:
            check['backendProtocol'] = utils.lookup(listener, 'defaultPool', 'protocol')
            check['backendPort'] = utils.lookup(listener, 'defaultPool', 'protocolPort')
            check['healthMonitorUuid'] = uuid
            check['interval'] = utils.lookup(listener, 'defaultPool', 'healthMonitor', 'interval')
            check['maxRetries'] = utils.lookup(listener, 'defaultPool', 'healthMonitor', 'maxRetries')
            check['timeout'] = utils.lookup(listener, 'defaultPool', 'healthMonitor', 'timeout')
            check['urlPath'] = utils.lookup(listener, 'defaultPool', 'healthMonitor', 'urlPath')

    if url and check['backendProtocol'] == 'TCP':
        raise exceptions.ArgumentError('--url cannot be used with TCP checks')

    # Update existing check with supplied values
    for key in clean_template.keys():
        check[key] = clean_template[key]

    try:
        mgr.update_lb_health_monitors(lb_uuid, [check])
        click.secho(f'Health Check {uuid} updated successfully', fg='green')
    except SoftLayerAPIError as exception:
        click.secho(f'Failed to update {uuid}', fg='red')
        click.secho(f"ERROR: {exception.faultString}", fg='red')



getAllObjects = [{
    'accountId': 123456,
    'createDate': '2020-09-15T13:12:08-06:00',
    'id': 112356450,
    'modifyDate': '2020-09-15T13:13:13-06:00',
    'status': 'COMPLETED',
    'userRecordId': 987456321,
    'userRecord': {
        'username': 'test@test.com'
    },
    'items': [
        {
            'categoryCode': 'port_speed',
            'description': '100 Mbps Private Network Uplink'
        },
        {
            'categoryCode': 'service_port',
            'description': '100 Mbps Private Uplink'
        },
        {
            'categoryCode': 'public_port',
            'description': '0 Mbps Public Uplink'
        }
    ],
    'orderApprovalDate': '2020-09-15T13:13:13-06:00',
    'orderTotalAmount': '0'
},
    {
        'accountId': 123456,
        'createDate': '2019-09-15T13:12:08-06:00',
        'id': 645698550,
        'modifyDate': '2019-09-15T13:13:13-06:00',
        'status': 'COMPLETED',
        'userRecordId': 987456321,
        'userRecord': {
            'username': 'test@test.com'
        },
        'items': [
            {
                'categoryCode': 'port_speed',
                'description': '100 Mbps Private Network Uplink'
            },

        ],
        'orderApprovalDate': '2019-09-15T13:13:13-06:00',
        'orderTotalAmount': '0'
}]

getObject = {
    'accountId': 1234,
    'createDate': '2020-09-23T16:22:30-06:00',
    'id': 6543210,
    'impersonatingUserRecordId': None,
    'initialInvoice': {
        'amount': '0',
        'id': 60012345,
        'invoiceTotalAmount': '0'
    },
    'items': [{
        'description': 'Dual Intel Xeon Silver 4210 (20 Cores, 2.20 GHz)'
    }],
    'modifyDate': '2020-09-23T16:22:32-06:00',
    'orderQuoteId': None,
    'orderTypeId': 11,
    'presaleEventId': None,
    'privateCloudOrderFlag': False,
    'status': 'APPROVED',
    'userRecord': {
        'displayName': 'testUser'
    },
    'userRecordId': 7654321,
}



getObject = {
    'accountId': 307608,
    'createDate': '2015-05-05T16:23:52-06:00',
    'id': 11449,
    'modifyDate': '2015-05-05T16:24:09-06:00',
    'name': 'TEST307608-1',
    'typeId': 2,
    'description': 'Citrix NetScaler VPX 10.5 10Mbps Standard',
    'managementIpAddress': '10.11.11.112',
    'outboundPublicBandwidthUsage': '.00365',
    'primaryIpAddress': '19.4.24.16',
    'datacenter': {
        'longName': 'Dallas 9',
        'name': 'dal09',
    },
    "networkVlans": [
        {
            "accountId": 11111,
            "id": 33333,
            "modifyDate": "2019-07-17T01:09:38+08:00",
            "name": "FirewallTesting",
            "primarySubnetId": 91111,
            "vlanNumber": 1711
        },
        {
            "accountId": 11111,
            "id": 862222,
            "modifyDate": "2019-07-17T01:09:42+08:00",
            "primarySubnetId": 502211,
            "vlanNumber": 722
        }
    ],
    'password': {
        'password': 'aaaaa',
        'username': 'root'
    },
    "subnets": [
        {
            "broadcastAddress": "",
            "cidr": 32,
            "gateway": "",
            "id": 74222,
            "modifyDate": "2016-10-26T23:39:12+08:00",
            "netmask": "255.255.255.255",
            "networkIdentifier": "159.253.111.111",
            "networkVlanId": 3611111,
            "sortOrder": "4",
            "subnetType": "STATIC_IP_ROUTED",
            "totalIpAddresses": "2",
            "usableIpAddressCount": "2",
            "version": 4,
            "addressSpace": "PUBLIC"
        }
    ],
    'type': {
        'keyName': 'NETSCALER_VPX',
        'name': 'NetScaler VPX'
    }
}



"""Adds an update to an existing ticket."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import helpers
from SoftLayer.CLI import ticket


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--body', help="Text to add to the ticket. STDIN or the default text editor will be used otherwise.")
@environment.pass_env
def cli(env, identifier, body):
    """Adds an update to an existing ticket.

Will update the ticket with `Some text`.::

        slcli ticket update 123456 --body="Some text"

Will update the ticket with text from STDIN::

        cat sometfile.txt | slcli ticket update 123456

Will open the default text editor, and once closed, use that text to update the ticket::

        slcli ticket update 123456
    """
    mgr = SoftLayer.TicketManager(env.client)

    ticket_id = helpers.resolve_id(mgr.resolve_ids, identifier, 'ticket')
    if body is None:
        stdin = click.get_text_stream('stdin')
        # Means there is text on the STDIN buffer, read it and add to the ticket
        if not stdin.isatty():
            body = stdin.read()
        # This is an interactive terminal, open a text editor
        else:
            body = click.edit('\n\n' + ticket.TEMPLATE_MSG)
    mgr.update_ticket(ticket_id=ticket_id, body=body)
    env.fout("Ticket Updated!")



"""Support tickets."""
import re

import click

from SoftLayer.CLI import formatting

TEMPLATE_MSG = "***** SoftLayer Ticket Content ******"

# https://softlayer.github.io/reference/services/SoftLayer_Ticket_Priority/getPriorities/
PRIORITY_MAP = [
    'No Priority',
    'Severity 1 - Critical Impact / Service Down',
    'Severity 2 - Significant Business Impact',
    'Severity 3 - Minor Business Impact',
    'Severity 4 - Minimal Business Impact'
]


def get_ticket_results(mgr, ticket_id, is_json=False, update_count=1):
    """Get output about a ticket.

    :param integer id: the ticket ID
    :param integer update_count: number of entries to retrieve from ticket
    :returns: a KeyValue table containing the details of the ticket

    """
    ticket = mgr.get_ticket(ticket_id)

    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'

    table.add_row(['id', ticket['id']])
    table.add_row(['Case_Number', ticket['serviceProviderResourceId']])
    table.add_row(['title', ticket['title']])
    table.add_row(['priority', PRIORITY_MAP[ticket.get('priority', 0)]])
    if ticket.get('assignedUser'):
        user = ticket['assignedUser']
        table.add_row([
            'user',
            "%s %s" % (user.get('firstName'), user.get('lastName')),
        ])

    table.add_row(['status', ticket['status']['name']])
    table.add_row(['created', ticket.get('createDate')])
    table.add_row(['edited', ticket.get('lastEditDate')])

    # Only show up to the specified update count
    updates = ticket.get('updates', [])
    count = min(len(updates), update_count)
    count_offset = len(updates) - count + 1  # Display as one-indexed
    for i, update in enumerate(updates[-count:]):
        wrapped_entry = ""

        # Add user details (fields are different between employee and users)
        editor = update.get('editor')
        if editor:
            if editor.get('displayName'):
                wrapped_entry += "By %s (Employee)\n" % (editor['displayName'])
            if editor.get('firstName'):
                wrapped_entry += "By %s %s\n" % (editor.get('firstName'),
                                                 editor.get('lastName'))

        # NOTE(kmcdonald): Windows new-line characters need to be stripped out
        wrapped_entry += click.wrap_text(update['entry'].replace('\r', ''))
        if is_json and '\n' in wrapped_entry:
            wrapped_entry = re.sub(r"(?<!\\)\n", " ", wrapped_entry)
        table.add_row(['update %s' % (count_offset + i,), wrapped_entry])
    return table



"""Manages Reserved Capacity."""
# :license: MIT, see LICENSE for more details.

import importlib
import os

import click
from SoftLayer.CLI.command import CommandLoader
from SoftLayer.CLI.command import OptionHighlighter

CONTEXT = {'help_option_names': ['-h', '--help'],
           'max_content_width': 999}


class CapacityCommands(CommandLoader):
    """Loads module for capacity related commands.

    Will automatically replace _ with - where appropriate.
    I'm not sure if this is better or worse than using a long list of manual routes, so I'm trying it here.
    CLI/virt/capacity/create_guest.py -> slcli vs capacity create-guest
    """

    def __init__(self, **attrs):
        CommandLoader.__init__(self, **attrs)
        self.path = os.path.dirname(__file__)
        self.highlighter = OptionHighlighter()
        self.env = None
        self.console = None

    def list_commands(self, ctx):
        """List all sub-commands."""
        commands = []
        for filename in os.listdir(self.path):
            if filename == '__init__.py':
                continue
            if filename.endswith('.py'):
                commands.append(filename[:-3].replace("_", "-"))
        commands.sort()
        return commands

    def get_command(self, ctx, cmd_name):
        """Get command for click."""
        path = "%s.%s" % (__name__, cmd_name)
        path = path.replace("-", "_")
        try:
            module = importlib.import_module(path)
            return getattr(module, 'cli')
        except ModuleNotFoundError as ex:
            print(ex.name)
            return None


# Required to get the sub-sub-sub command to work.
@click.group(cls=CapacityCommands, context_settings=CONTEXT)
def cli():
    """Base command for all capacity related concerns"""



"""List Reserved Capacity"""

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers.vs_capacity import CapacityManager as CapacityManager


@click.command(cls=SLCommand)
@environment.pass_env
def cli(env):
    """List Reserved Capacity groups."""
    manager = CapacityManager(env.client)
    result = manager.list()
    table = formatting.Table(
        ["ID", "Name", "Capacity", "Flavor", "Location", "Created"],
        title="Reserved Capacity"
    )
    for r_c in result:
        occupied_string = "#" * int(r_c.get('occupiedInstanceCount', 0))
        available_string = "-" * int(r_c.get('availableInstanceCount', 0))

        try:
            flavor = r_c['instances'][0]['billingItem']['description']
            # cost = float(r_c['instances'][0]['billingItem']['hourlyRecurringFee'])
        except KeyError:
            flavor = "Unknown Billing Item"
        location = r_c['backendRouter']['hostname']
        capacity = f"{occupied_string}{available_string}"
        table.add_row([r_c['id'], r_c['name'], capacity, flavor, location, r_c['createDate']])
    env.fout(table)



"""List Layer7 policies"""
import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.option('--protocol-id', '-p',
              required=False,
              type=int,
              help="Front-end Protocol identifier")
@environment.pass_env
def policies(env, protocol_id):
    """List policies of the front-end protocol (listener).

    Example::

        slcli loadbal l7policies (--protocol-id PROTOCOL_ID)
    """
    mgr = SoftLayer.LoadBalancerManager(env.client)

    if protocol_id:
        l7policies = mgr.get_l7policies(protocol_id)
        table = generate_l7policies_table(l7policies, protocol_id)
    else:
        l7policies = mgr.get_all_l7policies()
        table = l7policies_table(l7policies)
    env.fout(table)


def generate_l7policies_table(l7policies, identifier):
    """Takes a list of Layer7 policies and makes a table"""
    table = formatting.Table([
        'Id', 'UUID', 'Name', 'Action', 'Redirect', 'Priority', 'Create Date'
    ], title=f"Layer7 policies - protocol ID {identifier}")

    table.align['Name'] = 'l'
    table.align['Action'] = 'l'
    table.align['Redirect'] = 'l'
    for l7policy in sorted(l7policies, key=lambda data: data.get('priority')):
        table.add_row([
            l7policy.get('id'),
            l7policy.get('uuid'),
            l7policy.get('name'),
            l7policy.get('action'),
            l7policy.get('redirectL7PoolId') or l7policy.get('redirectUrl') or formatting.blank(),
            l7policy.get('priority'),
            l7policy.get('createDate'),
        ])
    return table


def l7policies_table(listeners):
    """Takes a dict of (protocols: policies list) and makes a list of tables"""
    tables = []
    for listener_id, list_policy in listeners.items():
        tables.append(generate_l7policies_table(list_policy, listener_id))
    return tables



"""List Subject IDs for ticket creation."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@environment.pass_env
def cli(env):
    """List Subject IDs for ticket creation."""
    ticket_mgr = SoftLayer.TicketManager(env.client)

    table = formatting.Table(['id', 'subject'])
    for subject in ticket_mgr.list_subjects():
        table.add_row([subject['id'], subject['name']])

    env.fout(table)



"""Prints out an SSH key to the screen."""
# :license: MIT, see LICENSE for more details.
from os import path

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--out-file', '-f',
              type=click.Path(exists=True, writable=True),
              help="The public SSH key will be written to this file")
@environment.pass_env
def cli(env, identifier, out_file):
    """Prints out an SSH key to the screen."""

    mgr = SoftLayer.SshKeyManager(env.client)

    key_id = helpers.resolve_id(mgr.resolve_ids, identifier, 'SshKey')

    key = mgr.get_key(key_id)

    if out_file:
        with open(path.expanduser(out_file), 'w', encoding="utf-8") as pub_file:
            pub_file.write(key['key'])

    table = formatting.KeyValueTable(['name', 'value'])
    table.add_row(['id', key['id']])
    table.add_row(['label', key.get('label')])
    table.add_row(['notes', key.get('notes', '-')])
    env.fout(table)



"""List package presets."""
# :license: MIT, see LICENSE for more details.

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers import ordering

COLUMNS = ['Name', 'Key Name', 'Description']


@click.command(cls=SLCommand)
@click.argument('package_keyname')
@click.option('--keyword', help="A word (or string) used to filter preset names.")
@click.option('--prices', '-p', is_flag=True, help='Use --prices to list the server item prices.')
@environment.pass_env
def cli(env, package_keyname, keyword, prices):
    """List package presets.

    .. Note::
        Presets are set CPU / RAM / Disk allotments. You still need to specify required items.
        Some packages do not have presets.
        Cost includes all items in a preset, and may include optional items.

    ::

        # List the presets for Bare Metal servers
        slcli order preset-list BARE_METAL_SERVER

        # List the Bare Metal server presets that include a GPU
        slcli order preset-list BARE_METAL_SERVER --keyword gpu

        # Get a specific flavor for Virtual Server
        slcli order preset-list PUBLIC_CLOUD_SERVER --prices --keyword BL2.56x242x

        # All packages with active presets
        slcli call-api SoftLayer_Product_Package getAllObjects --mask="mask[id,keyName,activePresetCount]" \
--json-filter='{"activePresets":{"operation":"not null"}}'
        ┌───────────────────┬──────┬──────────────────────────────────────────────────────────────────┐
        │ activePresetCount │  id  │                             keyName                              │
        ├───────────────────┼──────┼──────────────────────────────────────────────────────────────────┤
        │         1         │ 144  │                              3U_GPU                              │
        │         6         │ 200  │                        BARE_METAL_SERVER                         │
        │         1         │ 571  │                           NETWORK_VLAN                           │
        │        100        │ 835  │                       PUBLIC_CLOUD_SERVER                        │
        │         6         │ 865  │                     NETWORK_VLAN_FOR_SERVICE                     │
        │         5         │ 885  │                           8U_BI_S2_H4                            │
        │        56         │ 991  │                      TRANSIENT_CLOUD_SERVER                      │
        │        56         │ 1035 │                       SUSPEND_CLOUD_SERVER                       │
        │         9         │ 1045 │                          2U_BI_S3_H2000                          │
        │         7         │ 1075 │                   2U_IC4V_FIXED_CONFIGURATIONS                   │
        │        32         │ 1109 │                           BI_S4_H2000                            │
        │         8         │ 1117 │                           BI_S4_H4000                            │
        │         7         │ 1119 │                           BI_S4_H8000                            │
        │         5         │ 2636 │                    2U_BI_S4_H2000_AEP_ENABLED                    │
        │         5         │ 2676 │                    4U_BI_S4_H4000_AEP_ENABLED                    │
        │         5         │ 2700 │                    4U_BI_S4_H8000_AEP_ENABLED                    │
        │         6         │ 2866 │ ORACLE_APPLICATION_CLUSTER_CASCADE_LAKE_SCALABLE_FAMILY_4_DRIVES │
        │         1         │ 2874 │          ORACLE_APPLICATION_CLUSTER_COFFEE_LAKE_E2174G           │
        └───────────────────┴──────┴──────────────────────────────────────────────────────────────────┘

    """
    click.secho("*NOTICE*: Cost includes all items in a preset, and may include optional items.", fg="yellow")
    manager = ordering.OrderingManager(env.client)

    _filter = {}
    if keyword:
        _filter = {'activePresets': {'name': {'operation': '*= %s' % keyword}}}
    presets = manager.list_presets(package_keyname, filter=_filter)

    if prices:
        table = formatting.Table(['Id', 'Key Name', 'Hourly', 'Monthly', 'Location'])
        table.align = {"Id": "center", "Key Name": "left", "Hourly": "right", "Monthly": "right", "Location": "center"}
        for price in presets:
            locations_list = []
            if price['locations'] != []:
                for location in price['locations']:
                    locations_list.append(location['name'])
                locations = ', '.join(locations_list)
            else:
                locations = "All"
            table.add_row([
                price['id'],
                price['keyName'],
                get_item_price_data(price['prices'], 'hourlyRecurringFee'),
                get_item_price_data(price['prices'], 'recurringFee'),
                locations
            ])

    else:
        table = formatting.Table(COLUMNS)
        for preset in presets:
            table.add_row([
                str(preset['name']).strip(),
                str(preset['keyName']).strip(),
                str(preset['description']).strip()
            ])

    env.fout(table)


def get_item_price_data(prices, item_attribute):
    """Given an SoftLayer_Product_Item_Price, returns its default price data"""
    result = 0.0
    for this_price in prices:
        if item_attribute in this_price:
            result = result + float(this_price[item_attribute])
    return round(result, 3)



"""Save a quote"""
# :license: MIT, see LICENSE for more details.
import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers import ordering
from SoftLayer.utils import clean_time


@click.command(cls=SLCommand)
@click.argument('quote')
@environment.pass_env
def cli(env, quote):
    """Save a quote"""

    manager = ordering.OrderingManager(env.client)
    result = manager.save_quote(quote)

    table = formatting.Table([
        'Id', 'Name', 'Created', 'Modified', 'Status'
    ])
    table.align['Name'] = 'l'

    table.add_row([
        result.get('id'),
        result.get('name'),
        clean_time(result.get('createDate')),
        clean_time(result.get('modifyDate')),
        result.get('status'),
    ])

    env.fout(table)



"""Order/create a dedicated Host."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting
from SoftLayer.CLI import template


@click.command(cls=SoftLayer.CLI.command.SLCommand,
               epilog="See 'slcli dedicatedhost create-options' for valid options.")
@click.option('--hostname', '-H',
              help="Host portion of the FQDN",
              required=True,
              prompt=True)
@click.option('--router', '-r',
              help="Router hostname ex. fcr02a.dal13",
              show_default=True)
@click.option('--domain', '-D',
              help="Domain portion of the FQDN",
              required=True,
              prompt=True)
@click.option('--datacenter', '-d', help="Datacenter shortname",
              required=True,
              prompt=True)
@click.option('--flavor', '-f', help="Dedicated Virtual Host flavor",
              required=True,
              prompt=True)
@click.option('--billing',
              type=click.Choice(['hourly', 'monthly']),
              default='hourly',
              show_default=True,
              help="Billing rate")
@click.option('--verify',
              is_flag=True,
              help="Verify dedicatedhost without creating it.")
@click.option('--template', '-t',
              is_eager=True,
              callback=template.TemplateCallback(list_args=['key']),
              help="A template file that defaults the command-line options",
              type=click.Path(exists=True, readable=True, resolve_path=True))
@click.option('--export',
              type=click.Path(writable=True, resolve_path=True),
              help="Exports options to a template file")
@environment.pass_env
def cli(env, **kwargs):
    """Order/create a dedicated host."""
    mgr = SoftLayer.DedicatedHostManager(env.client)

    order = {
        'hostname': kwargs['hostname'],
        'domain': kwargs['domain'],
        'flavor': kwargs['flavor'],
        'location': kwargs['datacenter'],
        'hourly': kwargs.get('billing') == 'hourly',
    }

    if kwargs['router']:
        order['router'] = kwargs['router']

    do_create = not (kwargs['export'] or kwargs['verify'])

    output = None

    result = mgr.verify_order(**order)
    table = formatting.Table(['Item', 'cost'])
    table.align['Item'] = 'r'
    table.align['cost'] = 'r'
    if len(result['prices']) != 1:
        raise exceptions.ArgumentError("More than 1 price was found or no "
                                       "prices found")
    price = result['prices']
    if order['hourly']:
        total = float(price[0].get('hourlyRecurringFee', 0.0))
    else:
        total = float(price[0].get('recurringFee', 0.0))

    if order['hourly']:
        table.add_row(['Total hourly cost', "%.2f" % total])
    else:
        table.add_row(['Total monthly cost', "%.2f" % total])

    output = []
    output.append(table)
    output.append(formatting.FormattedItem(
        '',
        ' -- ! Prices reflected here are retail and do not '
        'take account level discounts and are not guaranteed.'))

    if kwargs['export']:
        export_file = kwargs.pop('export')
        template.export_to_template(export_file, kwargs,
                                    exclude=['wait', 'verify'])
        env.fout('Successfully exported options to a template file.')

    if do_create:
        if not env.skip_confirmations and not formatting.confirm(
                "This action will incur charges on your account. "
                "Continue?"):
            raise exceptions.CLIAbort('Aborting dedicated host order.')

        result = mgr.place_order(**order)

        table = formatting.KeyValueTable(['name', 'value'])
        table.align['name'] = 'r'
        table.align['value'] = 'l'
        table.add_row(['id', result['orderId']])
        table.add_row(['created', result['orderDate']])
        output.append(table)

    env.fout(output)



"""Edit a virtual server's details."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--domain', '-D', help="Domain portion of the FQDN")
@click.option('--hostname', '-H',
              help="Host portion of the FQDN. example: server")
@click.option('--tag', '-g',
              multiple=True,
              help="Tags to set or empty string to remove all")
@click.option('--userdata', '-u', help="User defined metadata string")
@click.option('--userfile', '-F',
              help="Read userdata from file",
              type=click.Path(exists=True, readable=True, resolve_path=True))
@click.option('--public-speed',
              help="Public port speed.",
              default=None,
              type=click.Choice(['0', '10', '100', '1000', '10000']))
@click.option('--private-speed',
              help="Private port speed.",
              default=None,
              type=click.Choice(['0', '10', '100', '1000', '10000']))
@environment.pass_env
def cli(env, identifier, domain, userfile, tag, hostname, userdata,
        public_speed, private_speed):
    """Edit a virtual server's details."""

    if userdata and userfile:
        raise exceptions.ArgumentError(
            '[-u | --userdata] not allowed with [-F | --userfile]')

    data = {}

    if userdata:
        data['userdata'] = userdata
    elif userfile:
        with open(userfile, 'r', encoding="utf-8") as userfile_obj:
            data['userdata'] = userfile_obj.read()

    data['hostname'] = hostname
    data['domain'] = domain

    if tag:
        data['tags'] = ','.join(tag)

    vsi = SoftLayer.VSManager(env.client)
    vs_id = helpers.resolve_id(vsi.resolve_ids, identifier, 'VS')

    if vsi.edit(vs_id, **data):
        for key, value in data.items():
            if value is not None:
                env.fout(f"The {key} of virtual server instance: {vs_id} was updated.")

    if public_speed is not None:
        if vsi.change_port_speed(vs_id, True, int(public_speed)):
            env.fout(f"The public speed of virtual server instance: {vs_id} was updated.")

    if private_speed is not None:
        if vsi.change_port_speed(vs_id, False, int(private_speed)):
            env.fout(f"The private speed of virtual server instance: {vs_id} was updated.")



"""List everything that could be tagged."""
# :license: MIT, see LICENSE for more details.

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers.tags import TagManager


@click.command(cls=SLCommand)
@environment.pass_env
def cli(env):
    """List everything that could be tagged."""

    tag_manager = TagManager(env.client)
    tag_types = tag_manager.get_all_tag_types()
    for tag_type in tag_types:
        title = f"{tag_type['description']} ({tag_type['keyName']})"
        table = formatting.Table(['Id', 'Name'], title=title)
        resources = tag_manager.taggable_by_type(tag_type['keyName'])
        for resource in resources:
            table.add_row([
                resource['resource']['id'],
                tag_manager.get_resource_name(resource['resource'], tag_type['keyName'])
            ])
        env.fout(table)



"""
    SoftLayer.exceptions
    ~~~~~~~~~~~~~~~~~~~~
    Exceptions used throughout the library

    :license: MIT, see LICENSE for more details.
"""
# pylint: disable=C0103


class SoftLayerError(Exception):
    """The base SoftLayer error."""


class Unauthenticated(SoftLayerError):
    """Unauthenticated."""


class IAMError(SoftLayerError):
    """Errors from iam.cloud.ibm.com"""

    def __init__(self, fault_code, fault_string, url=None):
        SoftLayerError.__init__(self, fault_string)
        self.faultCode = fault_code
        self.faultString = fault_string
        self.url = url

    def __repr__(self):
        return f"{self.url} ({self.faultCode}): {self.faultString}"

    def __str__(self):
        return f"{self.url} ({self.faultCode}): {self.faultString}"


class SoftLayerAPIError(SoftLayerError):
    """SoftLayerAPIError is an exception raised during API errors.

    Provides faultCode and faultString properties.
    """

    def __init__(self, fault_code, fault_string, *args):
        SoftLayerError.__init__(self, fault_string, *args)
        self.faultCode = fault_code
        self.reason = self.faultString = fault_string

    def __repr__(self):
        return '<%s(%s): %s>' % (self.__class__.__name__, self.faultCode, self.faultString)

    def __str__(self):
        return '%s(%s): %s' % (self.__class__.__name__, self.faultCode, self.faultString)


class ParseError(SoftLayerAPIError):
    """Parse Error."""


class ServerError(SoftLayerAPIError):
    """Server Error."""


class ApplicationError(SoftLayerAPIError):
    """Application Error."""


class RemoteSystemError(SoftLayerAPIError):
    """System Error."""


class TransportError(SoftLayerAPIError):
    """Transport Error."""


# XMLRPC Errors
class NotWellFormed(ParseError):
    """Request was not well formed."""


class UnsupportedEncoding(ParseError):
    """Encoding not supported."""


class InvalidCharacter(ParseError):
    """There was an invalid character."""


class SpecViolation(ServerError):
    """There was a spec violation."""


class MethodNotFound(SoftLayerAPIError):
    """Method name not found."""


class InvalidMethodParameters(SoftLayerAPIError):
    """Invalid method paramters."""


class InternalError(ServerError):
    """Internal Server Error."""



"""Create a password for a software component"""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--username', '-U', required=True, help="The username part of the username/password pair")
@click.option('--password', '-P', required=True, help="The password part of the username/password pair.")
@click.option('--notes', '-n', help="A note string stored for this username/password pair.")
@click.option('--software', required=True, help="The name of this specific piece of software.")
@environment.pass_env
def cli(env, identifier, username, password, notes, software):
    """Create a password for a software component."""

    mgr = SoftLayer.HardwareManager(env.client)

    software_components = mgr.get_software_components(identifier)
    software_id = ''
    try:
        for software_component in software_components:
            if software_component['softwareLicense']['softwareDescription']['name'].lower() == software.lower():
                software_id = software_component['id']
    except KeyError as ex:
        raise exceptions.CLIAbort('Software not found') from ex

    template = {
        "notes": notes,
        "password": password,
        "softwareId": software_id,
        "username": username
    }

    result = mgr.create_credential(template)

    table = formatting.KeyValueTable(['Name', 'Value'])
    table.align['Name'] = 'r'
    table.align['Value'] = 'l'
    table.add_row(['Software Credential Id', result['id']])
    table.add_row(['Created', result['createDate']])
    table.add_row(['Username', result['username']])
    table.add_row(['Password', result['password']])
    table.add_row(['Notes', result.get('notes') or formatting.blank()])
    env.fout(table)



getObject = {
    "id": 11111,
    "allowedVirtualGuests": [
        {
            "id": 22222,
            "allowedHost": {
                "accountId": 12345,
                "id": 18311111,
                "name": "iqn.2020-03.com.ibm:sl02su11111-v62941551",
                "resourceTableId": 6222222,
                "resourceTableName": "VIRTUAL_GUEST",
                "credential": {
                    "accountId": "12345",
                    "createDate": "2020-03-20T13:35:47-06:00",
                    "id": 1522222,
                    "nasCredentialTypeId": 2,
                    "password": "SjFDCpHrmKewos",
                    "username": "SL02SU322222-V62922222"
                }
            }
        }
    ]
}



"""
    SoftLayer.bandwidth
    ~~~~~~~~~~~~~~~~~~~~~~~
    Bandwidth manager

    :license: MIT, see License for more details.
"""

from SoftLayer import utils

# pylint: disable=invalid-name


class BandwidthManager(utils.IdentifierMixin, object):
    """Common functions for getting information from the Bandwidth service

    :param SoftLayer.API.BaseClient client: the client instance
    """

    def __init__(self, client):
        self.client = client

    def get_location_group(self):
        """Gets some basic bandwidth information

        :return: Bandwidth object
        """
        _filters = {"locationGroupTypeId": {"operation": 1}}

        return self.client.call('Location_Group', 'getAllObjects', filter=_filters)

    def create_pool(self, name_pool, id_location_group):
        """Gets some basic bandwidth information

        :return: Bandwidth object
        """
        current_user = self.client.call('Account', 'getCurrentUser')
        template = {
            "accountId": current_user['accountId'],
            "bandwidthAllotmentTypeId": 2,
            "locationGroupId": id_location_group,
            "name": name_pool
        }

        return self.client.call('Network_Bandwidth_Version1_Allotment', 'createObject', template)

    def get_bandwidth_detail(self, identifier):
        """Gets bandwidth pool detail.

        :returns: bandwidth pool detail
        """
        _mask = """activeDetails[allocation],projectedPublicBandwidthUsage, billingCyclePublicBandwidthUsage,
        hardware[outboundBandwidthUsage,bandwidthAllotmentDetail[allocation]],inboundPublicBandwidthUsage,
        virtualGuests[outboundPublicBandwidthUsage,bandwidthAllotmentDetail[allocation]],
        bareMetalInstances[outboundBandwidthUsage,bandwidthAllotmentDetail[allocation]]"""
        return self.client['SoftLayer_Network_Bandwidth_Version1_Allotment'].getObject(id=identifier, mask=_mask)

    def edit_pool(self, identifier, new_name_pool):
        """Edit bandwidth pool name

        :return: Bandwidth object
        """
        actual_bandwidth = self.get_bandwidth_detail(identifier)
        template = {
            "accountId": actual_bandwidth.get('accountId'),
            "bandwidthAllotmentTypeId": actual_bandwidth.get('bandwidthAllotmentTypeId'),
            "locationGroupId": actual_bandwidth.get('locationGroupId'),
            "name": new_name_pool
        }

        return self.client.call('Network_Bandwidth_Version1_Allotment', 'editObject', template, id=identifier)

    def delete_pool(self, identifier):
        """Delete bandwidth pool

        :return: Boolean value
        """
        return self.client.call('Network_Bandwidth_Version1_Allotment', 'requestVdrCancellation', id=identifier)



"""
    SoftLayer.metadata
    ~~~~~~~~~~~~~~~~~~
    Metadata Manager/helpers

    :license: MIT, see LICENSE for more details.
"""
from SoftLayer.API import BaseClient
from SoftLayer import consts
from SoftLayer import exceptions
from SoftLayer import transports


METADATA_MAPPING = {
    'backend_mac': {'call': 'getBackendMacAddresses'},
    'datacenter': {'call': 'getDatacenter'},
    'datacenter_id': {'call': 'getDatacenterId'},
    'domain': {'call': 'getDomain'},
    'frontend_mac': {'call': 'getFrontendMacAddresses'},
    'fqdn': {'call': 'getFullyQualifiedDomainName'},
    'hostname': {'call': 'getHostname'},
    'id': {'call': 'getId'},
    'primary_backend_ip': {'call': 'getPrimaryBackendIpAddress'},
    'primary_ip': {'call': 'getPrimaryIpAddress'},
    'primary_frontend_ip': {'call': 'getPrimaryIpAddress'},
    'provision_state': {'call': 'getProvisionState'},
    'router': {'call': 'getRouter', 'param_req': True},
    'tags': {'call': 'getTags'},
    'user_data': {'call': 'getUserMetadata'},
    'user_metadata': {'call': 'getUserMetadata'},
    'vlan_ids': {'call': 'getVlanIds', 'param_req': True},
    'vlans': {'call': 'getVlans', 'param_req': True},
}
METADATA_ATTRIBUTES = METADATA_MAPPING.keys()


class MetadataManager(object):
    """Provides an interface for the SoftLayer metadata service.

    See product information here:
    http://sldn.softlayer.com/reference/services/SoftLayer_Resource_Metadata

    This provides metadata about the resourse it is called from.
    See `METADATA_ATTRIBUTES` for full list of attributes.

        Usage:

            >>> import SoftLayer
            >>> client = SoftLayer.create_client_from_env()
            >>> from SoftLayer import MetadataManager
            >>> meta = MetadataManager(client)
            >>> meta.get('datacenter')
            'dal05'
            >>> meta.get('fqdn')
            'test.example.com'

    :param SoftLayer.API.BaseClient client: the client instance

    """

    attribs = METADATA_MAPPING

    def __init__(self, client=None, timeout=5):
        if client is None:
            transport = transports.RestTransport(
                timeout=timeout,
                endpoint_url=consts.API_PRIVATE_ENDPOINT_REST,
            )
            client = BaseClient(transport=transport)

        self.client = client

    def get(self, name, param=None):
        """Retreive a metadata attribute.

        :param string name: name of the attribute to retrieve. See `attribs`
        :param param: Required parameter for some attributes

        """
        if name not in self.attribs:
            raise exceptions.SoftLayerError('Unknown metadata attribute.')

        call_details = self.attribs[name]

        if call_details.get('param_req'):
            if not param:
                raise exceptions.SoftLayerError(
                    'Parameter required to get this attribute.')

        params = tuple()
        if param is not None:
            params = (param,)
        try:
            return self.client.call('Resource_Metadata',
                                    self.attribs[name]['call'],
                                    *params)
        except exceptions.SoftLayerAPIError as ex:
            if ex.faultCode == 404:
                return None
            raise ex

    def _get_network(self, kind, router=True, vlans=True, vlan_ids=True):
        """Wrapper for getting details about networks.

            :param string kind: network kind. Typically 'public' or 'private'
            :param boolean router: flag to include router information
            :param boolean vlans: flag to include vlan information
            :param boolean vlan_ids: flag to include vlan_ids

        """
        network = {}
        macs = self.get('%s_mac' % kind)
        network['mac_addresses'] = macs

        if len(macs) == 0:
            return network

        if router:
            network['router'] = self.get('router', macs[0])

        if vlans:
            network['vlans'] = self.get('vlans', macs[0])

        if vlan_ids:
            network['vlan_ids'] = self.get('vlan_ids', macs[0])

        return network

    def public_network(self, **kwargs):
        """Returns details about the public network.

        :param boolean router: True to return router details
        :param boolean vlans: True to return vlan details
        :param boolean vlan_ids: True to return vlan_ids

        """
        return self._get_network('frontend', **kwargs)

    def private_network(self, **kwargs):
        """Returns details about the private network.

        :param boolean router: True to return router details
        :param boolean vlans: True to return vlan details
        :param boolean vlan_ids: True to return vlan_ids

        """
        return self._get_network('backend', **kwargs)



"""Remove VLANs trunked to this server."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('hardware', nargs=1)
@click.argument('vlans', nargs=-1)
@click.option('--all', 'all_vlans', is_flag=True, default=False, help="Remove ALL trunked vlans from this server.")
@environment.pass_env
def cli(env, hardware, vlans, all_vlans):
    """Remove VLANs trunked to this server.

    HARDWARE is the id of the server
    VLANS is the ID, name, or number of the VLANs you want to remove. Multiple vlans can be removed at the same time.
    It is recommended to use the vlan ID, especially if you have multiple vlans with the same name/number.
    """
    if not vlans and not all_vlans:
        raise exceptions.ArgumentError("Error: Missing argument 'VLANS'.")
    h_mgr = SoftLayer.HardwareManager(env.client)
    n_mgr = SoftLayer.NetworkManager(env.client)
    hw_id = helpers.resolve_id(h_mgr.resolve_ids, hardware, 'hardware')

    if all_vlans:
        h_mgr.clear_vlan(hw_id)
        env.fout("Done.")
        return

    # Enclosing in quotes is required for any input that has a space in it.
    # "Public DAL10" for example needs to be sent to search as \"Public DAL10\"
    sl_vlans = n_mgr.search_for_vlan(" ".join(f"\"{v}\"" for v in vlans))
    if not sl_vlans:
        raise exceptions.ArgumentError(f"No vlans found matching {' '.join(vlans)}")
    del_vlans = parse_vlans(sl_vlans)
    component_mask = "mask[id, name, port, macAddress, primaryIpAddress]"
    # NEXT: Add nice output / exception handling
    if len(del_vlans['public']) > 0:
        components = h_mgr.get_network_components(hw_id, mask=component_mask, space='public')
        for c in components:
            if c.get('primaryIpAddress'):
                h_mgr.remove_vlan(c.get('id'), del_vlans['public'])
    if len(del_vlans['private']) > 0:
        components = h_mgr.get_network_components(hw_id, mask=component_mask, space='private')
        for c in components:
            if c.get('primaryIpAddress'):
                h_mgr.remove_vlan(c.get('id'), del_vlans['private'])


def parse_vlans(vlans):
    """returns a dictionary mapping for public / private vlans"""

    pub_vlan = []
    pri_vlan = []
    for vlan in vlans:
        if vlan.get('networkSpace') == "PUBLIC":
            pub_vlan.append(vlan)
        else:
            pri_vlan.append(vlan)
    return {"public": pub_vlan, "private": pri_vlan}



"""Get all hardware notifications associated with the passed hardware ID."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Get all hardware notifications."""

    hardware = SoftLayer.HardwareManager(env.client)

    notifications = hardware.get_notifications(identifier)

    table = formatting.KeyValueTable(['Id', 'Domain', 'Hostmane', 'Username', 'Email', 'FirstName', 'Lastname'])
    table.align['Domain'] = 'r'
    table.align['Username'] = 'l'

    for notification in notifications:
        table.add_row([notification['id'],
                       notification['hardware']['fullyQualifiedDomainName'], notification['hardware']['hostname'],
                       notification['user']['username'], notification['user']['email'],
                       notification['user']['firstName'], notification['user']['lastName']])

    env.fout(table)



getObject = {
    'id': 37401,
    'memoryCapacity': 242,
    'modifyDate': '',
    'name': 'test-dedicated',
    'diskCapacity': 1200,
    'createDate': '2017-10-16T12:50:23-05:00',
    'cpuCount': 56,
    'accountId': 1199911
}


getAvailableRouters = [
    {'hostname': 'bcr01a.dal05', 'id': 12345},
    {'hostname': 'bcr02a.dal05', 'id': 12346},
    {'hostname': 'bcr03a.dal05', 'id': 12347},
    {'hostname': 'bcr04a.dal05', 'id': 12348}
]

getObjectById = {
    'datacenter': {
        'id': 12345,
        'name': 'dal05',
        'longName': 'Dallas 5'
    },
    'memoryCapacity': 242,
    'modifyDate': '2017-11-06T11:38:20-06:00',
    'name': 'test-dedicated',
    'diskCapacity': 1200,
    'backendRouter': {
        'domain': 'test.com',
        'hostname': 'bcr01a.dal05',
        'id': 12345
    },
    'guestCount': 1,
    'cpuCount': 56,
    'guests': [{
        'domain': 'test.com',
        'hostname': 'test-dedicated',
        'id': 12345,
        'uuid': 'F9329795-4220-4B0A-B970-C86B950667FA'
    }],
    'billingItem': {
        'nextInvoiceTotalRecurringAmount': 1515.556,
        'orderItem': {
            'id': 12345,
            'order': {
                'status': 'APPROVED',
                'privateCloudOrderFlag': False,
                'modifyDate': '2017-11-02T11:42:50-07:00',
                'orderQuoteId': '',
                'userRecordId': 12345,
                'createDate': '2017-11-02T11:40:56-07:00',
                'impersonatingUserRecordId': '',
                'orderTypeId': 7,
                'presaleEventId': '',
                'userRecord': {
                    'username': 'test-dedicated'
                },
                'id': 12345,
                'accountId': 12345
            }
        },
        'id': 12345,
        'children': [
            {
                'nextInvoiceTotalRecurringAmount': 0.0,
                'categoryCode': 'dedicated_host_ram'
            },
            {
                'nextInvoiceTotalRecurringAmount': 0.0,
                'categoryCode': 'dedicated_host_disk'
            }
        ]
    },
    'id': 12345,
    'createDate': '2017-11-02T11:40:56-07:00'
}

deleteObject = True

getGuests = [{
    'id': 200,
    'hostname': 'vs-test1',
    'domain': 'test.sftlyr.ws',
    'fullyQualifiedDomainName': 'vs-test1.test.sftlyr.ws',
    'status': {'keyName': 'ACTIVE', 'name': 'Active'},
    'datacenter': {'id': 50, 'name': 'TEST00',
                   'description': 'Test Data Center'},
    'powerState': {'keyName': 'RUNNING', 'name': 'Running'},
    'maxCpu': 2,
    'maxMemory': 1024,
    'primaryIpAddress': '172.16.240.2',
    'globalIdentifier': '1a2b3c-1701',
    'primaryBackendIpAddress': '10.45.19.37',
    'hourlyBillingFlag': False,
    'billingItem': {
        'id': 6327,
        'recurringFee': 1.54,
        'orderItem': {
            'order': {
                'userRecord': {
                    'username': 'chechu',
                }
            }
        }
    },
}, {
    'id': 202,
    'hostname': 'vs-test2',
    'domain': 'test.sftlyr.ws',
    'fullyQualifiedDomainName': 'vs-test2.test.sftlyr.ws',
    'status': {'keyName': 'ACTIVE', 'name': 'Active'},
    'datacenter': {'id': 50, 'name': 'TEST00',
                   'description': 'Test Data Center'},
    'powerState': {'keyName': 'RUNNING', 'name': 'Running'},
    'maxCpu': 4,
    'maxMemory': 4096,
    'primaryIpAddress': '172.16.240.7',
    'globalIdentifier': '05a8ac-6abf0',
    'primaryBackendIpAddress': '10.45.19.35',
    'hourlyBillingFlag': True,
    'billingItem': {
        'id': 6327,
        'recurringFee': 1.54,
        'orderItem': {
            'order': {
                'userRecord': {
                    'username': 'chechu',
                }
            }
        }
    }
}]



findByHardwareId = [
    {
        'hardwareId': 147258,
        'id': 1232569,
        'userId': 3698,
        'hardware': {
            'accountId': 963258,
            'domain': 'testedit.com',
            'fullyQualifiedDomainName': 'testslcli.testedit.com',
            'hostname': 'bardcabero',
            'id': 1403539,
            'notes': 'My golang note',
            'provisionDate': '2020-04-27T16:10:56-06:00',
        },
        'user': {
            'accountId': 307608,
            'email': 'cgallo@us.ibm.com',
            'firstName': 'CHRISTOPHER',
            'id': 167758,
            'lastName': 'GALLO',
            'username': 'SL307608',
        }
    },
    {
        'hardwareId': 1403539,
        'id': 1408587,
        'userId': 9734826,
        'hardware': {
            'accountId': 963258,
            'domain': 'testedit.com',
            'fullyQualifiedDomainName': 'testslcli.testedit.com',
            'hostname': 'bardcabero',
            'id': 1403539,
            'notes': 'My golang note',
            'provisionDate': '2020-04-27T16:10:56-06:00',
        },
        'user': {
            'accountId': 307608,
            'email': 'Brian.Flores@ibm.com',
            'firstName': 'Brian',
            'id': 9734826,
            'lastName': 'Flores',
            'username': '307608_brian.flores@ibm.com',
        }
    }
]

createObject = {
    'hardwareId': 1403539,
    'id': 1408593,
    'userId': 7650493,
    'hardware': {
        'accountId': 307608,
        'domain': 'testedit.com',
        'fullyQualifiedDomainName': 'bardcabero.testedit.com',
        'hostname': 'bardcabero',
        'id': 1403539,
        'notes': 'My golang note',

    },
    'user': {
        'accountId': 307608,
        'email': 'daniel.cabero@jalasoft.com',
        'firstName': 'daniel',
        'id': 7650493,
        'lastName': 'cabero',
        'username': 'sl307608-dcabero'

    }
}

deleteObjects = True



"""Upgrade a virtual server."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers
from SoftLayer.CLI import virt


@click.command(cls=SoftLayer.CLI.command.SLCommand, epilog="""Note: SoftLayer automatically reboots the VS once
upgrade request is placed. The VS is halted until the Upgrade transaction is
completed. However for Network, no reboot is required.""")
@click.argument('identifier')
@click.option('--cpu', type=click.INT, help="Number of CPU cores")
@click.option('--private', is_flag=True,
              help="CPU core will be on a dedicated host server.")
@click.option('--memory', type=virt.MEM_TYPE, help="Memory in megabytes")
@click.option('--network', type=click.INT, help="Network port speed in Mbps")
@click.option('--add-disk', type=click.INT, multiple=True, required=False, help="add Hard disk in GB")
@click.option('--resize-disk', nargs=2, multiple=True, type=(int, int),
              help="Update disk number to size in GB. --resize-disk 250 2 ")
@click.option('--flavor', type=click.STRING,
              help="Flavor keyName\nDo not use --memory, --cpu or --private, if you are using flavors")
@environment.pass_env
def cli(env, identifier, cpu, private, memory, network, flavor, add_disk, resize_disk):
    """Upgrade a virtual server."""

    vsi = SoftLayer.VSManager(env.client)

    if not any([cpu, memory, network, flavor, resize_disk, add_disk]):
        raise exceptions.ArgumentError("Must provide [--cpu],"
                                       " [--memory], [--network], [--flavor], [--resize-disk], or [--add] to upgrade")

    if private and not cpu:
        raise exceptions.ArgumentError("Must specify [--cpu] when using [--private]")

    vs_id = helpers.resolve_id(vsi.resolve_ids, identifier, 'VS')
    if not (env.skip_confirmations or formatting.confirm("This action will incur charges on your account. Continue?")):
        raise exceptions.CLIAbort('Aborted')

    disk_json = []
    if memory:
        memory = int(memory / 1024)
    if resize_disk:
        for guest_disk in resize_disk:
            disks = {'capacity': guest_disk[0], 'number': guest_disk[1]}
            disk_json.append(disks)

    elif add_disk:
        for guest_disk in add_disk:
            disks = {'capacity': guest_disk, 'number': -1}
            disk_json.append(disks)

    if not vsi.upgrade(vs_id, cpus=cpu, memory=memory, nic_speed=network, public=not private, preset=flavor,
                       disk=disk_json):
        raise exceptions.CLIAbort('VS Upgrade Failed')



"""Displays information about the bandwidth pools"""
# :license: MIT, see LICENSE for more details.
import concurrent.futures as cf
import logging
import time

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers.account import AccountManager as AccountManager
from SoftLayer import utils

LOGGER = logging.getLogger(__name__)


@click.command(cls=SLCommand, )
@environment.pass_env
def cli(env):
    """Displays bandwidth pool information

    Similiar to https://cloud.ibm.com/classic-bandwidth/pools

    More information
    https://cloud.ibm.com/docs/bandwidth-metering?topic=bandwidth-metering-get-started-with-bandwidth-metering
    """

    manager = AccountManager(env.client)

    items = manager.get_bandwidth_pools()

    table = formatting.Table([
        "Id",
        "Name",
        "Region",
        "Devices",
        "Allocation",
        "Current Usage",
        "Projected Usage",
        "Cost",
        "Deletion"
    ], title="Bandwidth Pools")
    table.align = 'l'

    start_m = time.perf_counter()

    with cf.ThreadPoolExecutor(max_workers=5) as executor:
        for item, servers in zip(items, executor.map(manager.get_bandwidth_pool_counts,
                                                     [item.get('id') for item in items])):

            id_bandwidth = item.get('id')
            name = item.get('name')
            region = utils.lookup(item, 'locationGroup', 'name')

            allocation = f"{item.get('totalBandwidthAllocated', 0)} GB"

            current = utils.lookup(item, 'billingCyclePublicBandwidthUsage', 'amountOut')
            if current is not None:
                current = f"{current} GB"
            else:
                current = "0 GB"

            projected = f"{item.get('projectedPublicBandwidthUsage', 0)} GB"

            cost = utils.lookup(item, 'billingItem', 'nextInvoiceTotalRecurringAmount')
            if cost is not None:
                cost = f"${cost}"
            else:
                cost = "$0.0"

            deletion = utils.clean_time(item.get('endDate'))
            if deletion == '':
                deletion = formatting.blank()

            table.add_row([id_bandwidth, name, region, servers, allocation, current, projected, cost, deletion])

    end_m = time.perf_counter()
    LOGGER.debug('Total API Call time %s', end_m - start_m)
    env.fout(table)



advancedSearch = [
    {
        "relevanceScore": "4",
        "resourceType": "SoftLayer_Hardware",
        "resource": {
            "accountId": 307608,
            "domain": "vmware.test.com",
            "fullyQualifiedDomainName": "host14.vmware.test.com",
            "hardwareStatusId": 5,
            "hostname": "host14",
            "id": 123456,
            "manufacturerSerialNumber": "AAAAAAAAA",
            "notes": "A test notes",
            "provisionDate": "2018-08-24T12:32:10-06:00",
            "serialNumber": "SL12345678",
            "serviceProviderId": 1,
            "hardwareStatus": {
                "id": 5,
                "status": "ACTIVE"
            }
        }
    }
]

search = advancedSearch

getObjectTypes = [{"name": "SoftLayer_Event_Log",
                   "properties": [
                       {
                           "name": "accountId",
                           "sortableFlag": True,
                           "type": "integer"
                       }]},
                  {"name": "SoftLayer_Hardware",
                   "properties": [
                       {
                           "name": "accountId",
                           "sortableFlag": True,
                           "type": "integer"
                       },
                       {
                           "name": "datacenter.longName",
                           "sortableFlag": True,
                           "type": "string"
                       },
                       {
                           "name": "deviceStatus.name",
                           "sortableFlag": True,
                           "type": "string"
                       }]
                   }]


advancedSearchVlan = [
    {
        "matchedTerms": [
            "name:|IBMPrivate|",
            "name.sort:|IBMPrivate|"
        ],
        "relevanceScore": "9.05",
        "resource": {
            "fullyQualifiedName": "dal10.bcr03.0000",
            "id": 11111,
            "name": "IBMPrivate",
            "vlanNumber": 0000,
            "networkSpace": "PRIVATE"
        },
        "resourceType": "SoftLayer_Network_Vlan"
    },
    {
        "matchedTerms": [
            "name:|IBMPublic|",
            "name.sort:|IBMPublic|"
        ],
        "relevanceScore": "9.01",
        "resource": {
            "fullyQualifiedName": "dal10.bcr03.11111",
            "id": 999999,
            "name": "IBMPublic",
            "vlanNumber": 11111,
            "networkSpace": "PUBLIC"
        },
        "resourceType": "SoftLayer_Network_Vlan"
    }
]



"""Capture virtual server image."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers


# pylint: disable=redefined-builtin

@click.command(cls=SoftLayer.CLI.command.SLCommand, short_help="Capture SoftLayer image.")
@click.argument('identifier')
@click.option('--name', '-n', required=True, help="Name of the image")
@click.option('--all', is_flag=True, default=False, help="Capture all disks belonging to the VS")
@click.option('--note', help="Add a note to be associated with the image")
@environment.pass_env
def cli(env, identifier, name, all, note):
    """Capture one or all disks from a virtual server to a SoftLayer image."""

    vsi = SoftLayer.VSManager(env.client)
    vs_id = helpers.resolve_id(vsi.resolve_ids, identifier, 'VS')

    capture = vsi.capture(vs_id, name, all, note)

    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'

    table.add_row(['globalIdentifier', capture['globalIdentifier']])
    table.add_row(['date', capture['createDate'][:10]])
    table.add_row(['time', capture['createDate'][11:19]])
    table.add_row(['transaction', formatting.transaction_status(capture)])
    table.add_row(['transaction_id', capture['id']])
    table.add_row(['all_disks', all])
    env.fout(table)



"""Vlan order options."""
# :license: MIT, see LICENSE for more details.
# pylint: disable=too-many-statements
import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, short_help="Get options to use for creating Vlan servers.")
@environment.pass_env
def cli(env):
    """List all the options for creating VLAN"""

    mgr = SoftLayer.NetworkManager(env.client)
    datacenters = mgr.get_list_datacenter()

    table = formatting.Table(['Options', 'Value'], title="Datacenters")
    router_table = formatting.Table(['Datacenter', 'Router/Pod'])
    dc_table = formatting.Table(['Datacenters'])
    table.add_row(['VLAN type', 'Private, Public'])

    for datacenter in datacenters:
        dc_table.add_row([datacenter['name']])
        routers = mgr.get_routers(datacenter['id'])
        for router in routers:
            router_table.add_row([datacenter['name'], router['hostname']])

    table.add_row(['Datacenters', dc_table])
    table.add_row(['Routers', router_table])

    env.fout(table)



getAllObjects = [
    {
        "id": 1001,
        "name": "Accounting Request"
    },
    {
        "id": 1002,
        "name": "Sales Request"
    },
    {
        "id": 1003,
        "name": "Reboots and Console Access"
    },
    {
        "id": 1004,
        "name": "DNS Request"
    },
    {
        "id": 1005,
        "name": "Hardware Issue"
    }
]



getObject = {
    'accountId': 12345,
    'address1': '315 Test Street',
    'apiAuthenticationKeys': [{'authenticationKey': 'aaaaaaaaaaaaaaaaaaaaaaaaa'}],
    'city': 'Houston',
    'companyName': 'SoftLayer Development Community',
    'country': 'US',
    'createDate': '2014-08-18T12:58:02-06:00',
    'displayName': 'Test',
    'email': 'test@us.ibm.com',
    'firstName': 'Test',
    'id': 244956,
    'isMasterUserFlag': False,
    'lastName': 'Testerson',
    'openIdConnectUserName': 'test@us.ibm.com',
    'parent': {
        'id': 167758, 'username': 'SL12345'},
    'parentId': 167758,
    'postalCode': '77002',
    'sslVpnAllowedFlag': True,
    'state': 'TX',
    'statusDate': None,
    'successfulLogins': [
        {'createDate': '2018-05-08T15:28:32-06:00',
         'ipAddress': '175.125.126.118',
         'successFlag': True,
         'userId': 244956},
    ],
    'timezone': {
        'id': 113,
        'longName': '(GMT-06:00) America/Chicago - CST',
        'name': 'America/Chicago',
        'offset': '-0600',
        'shortName': 'CST'},
    'timezoneId': 113,
    'unsuccessfulLogins': [
        {'createDate': '2018-02-09T14:13:15-06:00',
         'ipAddress': '73.136.219.36',
         'successFlag': False,
         'userId': 244956},
    ],
    'userStatus': {'name': 'Active'},
    'userStatusId': 1001,
    'username': 'SL12345-test',
    'vpnManualConfig': False,
    'permissions': [
        {'key': 'ALL_1',
         'keyName': 'ACCESS_ALL_HARDWARE',
         'name': 'All Hardware Access'}
    ],
    'roles': []
}

getPermissions = [
    {'key': 'ALL_1',
     'keyName': 'ACCESS_ALL_HARDWARE',
     'name': 'All Hardware Access'},
    {'key': 'A_1',
     'keyName': 'ACCOUNT_SUMMARY_VIEW',
     'name': 'View Account Summary'},
    {'key': 'A_10',
     'keyName': 'ADD_SERVICE_STORAGE',
     'name': 'Add/Upgrade Storage (StorageLayer)'}
]

getLoginAttempts = [
    {
        "createDate": "2017-10-03T09:28:33-06:00",
        "ipAddress": "1.2.3.4",
        "successFlag": False,
        "userId": 1111,
        "username": "sl1234"
    }
]

getOverrides = [
    {
        'id': 3661234,
        'subnetId': 1234
    }
]

addBulkPortalPermission = True
removeBulkPortalPermission = True
createObject = getObject
editObject = True
addApiAuthenticationKey = "1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a"
removeApiAuthenticationKey = True
updateVpnUser = True
removeDedicatedHostAccess = True
removeHardwareAccess = True
removeVirtualGuestAccess = True
addDedicatedHostAccess = True
addHardwareAccess = True
addVirtualGuestAccess = True
updateVpnPassword = True

getApiAuthenticationKeys = [
    {
        "authenticationKey": "1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a1a",
        "id": 123456,
        "timestampKey": 12121212,
        "userId": 111111
    }
]

getEmptyApiAuthenticationKeys = []

getHardware = [{
        "domain": "testedit.com",
        "fullyQualifiedDomainName": "test.testedit.com",
        "hardwareStatusId": 5,
        "hostname": "test",
        "id": 1403539,
        "manufacturerSerialNumber": "J33H9TX",
        "notes": "My golang note",
        "provisionDate": "2020-04-27T16:10:56-06:00",
        "serialNumber": "SL01FJUI",
        "globalIdentifier": "81434794-af69-44d5-bb97-6b6f43454eee",
        "hardwareStatus": {
            "id": 5,
            "status": "ACTIVE"
        },
        "networkManagementIpAddress": "10.93.138.222",
        "primaryBackendIpAddress": "10.93.138.202",
        "primaryIpAddress": "169.48.191.244",
        "privateIpAddress": "10.93.138.202"
    }]
getDedicatedHosts = [
    {
        'createDate': '2021-11-18T15:13:57-06:00',
        'diskCapacity': 1200,
        'id': 656700,
        'memoryCapacity': 242,
        'modifyDate': '2022-04-26T10:49:48-06:00',
        'name': 'dedicatedhost01'
    },
    {
        'accountId': 307608,
        'cpuCount': 56,
        'createDate': '2022-02-18T12:47:30-06:00',
        'diskCapacity': 1200,
        'id': 691394,
        'memoryCapacity': 242,
        'modifyDate': '2022-04-18T11:24:20-06:00',
        'name': 'test'
    }
]
getVirtualGuests = [
    {
        "fullyQualifiedDomainName": "KVM-Test.cgallo.com",
        "hostname": "KVM-Test",
        "id": 121401696,
        "maxCpu": 2,
        "maxCpuUnits": "CORE",
        "maxMemory": 4096,
        "modifyDate": "2022-01-25T23:23:13-06:00",
        "provisionDate": "2021-06-09T14:51:54-06:00",
        "startCpus": 2,
        "typeId": 1,
        "uuid": "15951561-6171-0dfc-f3d2-be039e51cc10",
        "globalIdentifier": "a245a7dd-acd1-4d1a-9356-cc1ac6b55b98",
        "primaryBackendIpAddress": "10.208.73.53",
        "primaryIpAddress": "169.48.96.27",
    },
    {
        "createDate": "2020-11-16T09:01:57-06:00",
        "deviceStatusId": 8,
        "domain": "softlayer.test",
        "fullyQualifiedDomainName": "SuspendVsTest.softlayer.test",
        "hostname": "SuspendVsTest",
        "id": 112238162,
        "maxCpu": 8,
        "maxCpuUnits": "CORE",
        "maxMemory": 16384,
        "modifyDate": "2022-01-25T14:15:37-06:00",
        "provisionDate": "2020-11-16T09:09:04-06:00",
        "startCpus": 8,
        "typeId": 1,
        "uuid": "d8908a64-f4d4-5637-49c7-650572d47120",
        "globalIdentifier": "7fe777af-d38b-47c2-9f1c-b1ec26751b58",
        "primaryBackendIpAddress": "10.74.54.76",
    }]



"""Remove block storage subnets for the given host id."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import environment


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('access_id', type=int)
@click.option('--subnet-id', multiple=True, type=int,
              help="ID of the subnets to remove; e.g.: --subnet-id 1234")
@environment.pass_env
def cli(env, access_id, subnet_id):
    """Remove block storage subnets for the given host id.

    Example::

                slcli block subnets-remove 111111 --subnet-id 222222
                slcli block subnets-remove 111111 --subnet-id 222222 --subnet-id 333333
                ACCESS_ID is the host_id obtained by: slcli block access-list <volume_id>

    access_id is the host_id obtained by: slcli block access-list <volume_id>

    SoftLayer_Account::iscsiisolationdisabled must be False to use this command
    """
    try:
        subnet_ids = list(subnet_id)
        block_manager = SoftLayer.BlockStorageManager(env.client)
        removed_subnets = block_manager.remove_subnets_from_acl(access_id, subnet_ids)

        for subnet in removed_subnets:
            message = f"Successfully removed subnet id: {subnet} for allowed host id: {access_id}"
            click.echo(message)

        failed_to_remove_subnets = list(set(subnet_ids) - set(removed_subnets))
        for subnet in failed_to_remove_subnets:
            message = f"Failed to remove subnet id: {subnet} for allowed host id: {access_id}"
            click.echo(message)

    except SoftLayer.SoftLayerAPIError as ex:
        message = f"Unable to remove subnets.\nReason: {ex.faultString}"
        click.echo(message)



"""Exit the shell."""
# :license: MIT, see LICENSE for more details.

import click

from SoftLayer.shell import core


@click.command()
def cli():
    """Exit the shell."""
    raise core.ShellExit()



"""Order/create a VLAN instance."""
# :license: MIT, see LICENSE for more details.
import click
import SoftLayer
from SoftLayer.managers import ordering

from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.option('--name', required=False, prompt=True, help="Vlan name")
@click.option('--datacenter', '-d', required=False, help="Datacenter shortname")
@click.option('--pod', '-p', required=False, help="Pod name. E.g dal05.pod01")
@click.option('--network', default='public', show_default=True, type=click.Choice(['public', 'private']),
              help='Network vlan type')
@click.option('--billing', default='hourly', show_default=True, type=click.Choice(['hourly', 'monthly']),
              help="Billing rate")
@environment.pass_env
def cli(env, name, datacenter, pod, network, billing):
    """Order/create a VLAN instance.

    Example:
        slcli vlan create --name myvlan --datacenter dal13
        or
        slcli vlan create --name myvlan --pod dal10.pod01
    """

    if not (datacenter or pod):
        raise exceptions.CLIAbort("--datacenter or --pod is required to create a VLAN")

    item_package = ['PUBLIC_NETWORK_VLAN']
    complex_type = 'SoftLayer_Container_Product_Order_Network_Vlan'
    extras = {'name': name}
    if pod:
        datacenter = pod.split('.')[0]
        mgr = SoftLayer.NetworkManager(env.client)
        pods = mgr.get_pods()
        for router in pods:
            if router.get('name') == pod:
                if network == 'public':
                    extras['routerId'] = router.get('frontendRouterId')
                elif network == 'private':
                    extras['routerId'] = router.get('backendRouterId')
                break
        if not extras.get('routerId'):
            raise exceptions.CLIAbort(f"Unable to find pod name: {pod}")
    if network == 'private':
        item_package = ['PRIVATE_NETWORK_VLAN']

    ordering_manager = ordering.OrderingManager(env.client)
    result = ordering_manager.place_order(package_keyname='NETWORK_VLAN',
                                          location=datacenter,
                                          item_keynames=item_package,
                                          complex_type=complex_type,
                                          hourly=billing,
                                          extras=extras)
    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'
    table.add_row(['id', result['orderId']])
    table.add_row(['created', result['orderDate']])
    table.add_row(['name', result['orderDetails']['orderContainers'][0]['name']])

    env.fout(table)



getObject = {
    "id": 405,
    "keyName": "AC1_8X60X25",
    "prices": [
        {
            "hourlyRecurringFee": "1.425",
            "id": 207345,
            "recurringFee": "936.23",
            "item": {
                "description": "1 x P100 GPU",
                "id": 10933,
                "keyName": "1_X_P100_GPU",
                "itemCategory": {
                    "categoryCode": "guest_pcie_device0",
                    "id": 1259
                }
            }
        },
        {
            "hourlyRecurringFee": "0",
            "id": 2202,
            "recurringFee": "0",
            "item": {
                "description": "25 GB (SAN)",
                "id": 1178,
                "keyName": "GUEST_DISK_25_GB_SAN",
                "itemCategory": {
                    "categoryCode": "guest_disk0",
                    "id": 81
                }
            }
        },
        {
            "hourlyRecurringFee": ".342",
            "id": 207361,
            "recurringFee": "224.69",
            "item": {
                "description": "60 GB",
                "id": 10939,
                "keyName": "RAM_0_UNIT_PLACEHOLDER_10",
                "itemCategory": {
                    "categoryCode": "ram",
                    "id": 3
                }
            }
        },
        {
            "hourlyRecurringFee": ".181",
            "id": 209595,
            "recurringFee": "118.26",
            "item": {
                "capacity": 8,
                "description": "8 x 2.0 GHz or higher Cores",
                "id": 11307,
                "keyName": "GUEST_CORE_8",
                "itemCategory": {
                    "categoryCode": "guest_core",
                    "id": 80
                }
            }
        }
    ]
}



"""
    SoftLayer Python API Client
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~
    SoftLayer API bindings

    Usage:

        >>> import SoftLayer
        >>> client = SoftLayer.create_client_from_env(username="username",
                                                      api_key="api_key")
        >>> resp = client.call('Account', 'getObject')
        >>> resp['companyName']
        'Your Company'

    :license: MIT, see LICENSE for more details.
"""
# pylint: disable=r0401,invalid-name,wildcard-import
# NOQA appears to no longer be working. The code might have been upgraded.
from SoftLayer import consts

from SoftLayer.API import *  # NOQA
from SoftLayer.managers import *  # NOQA
from SoftLayer.exceptions import *  # NOQA
from SoftLayer.auth import *  # NOQA
from SoftLayer.transports import *  # NOQA

__title__ = 'SoftLayer'
__version__ = consts.VERSION
__author__ = 'SoftLayer Technologies, Inc.'
__license__ = 'MIT'
__copyright__ = 'Copyright 2016 SoftLayer Technologies, Inc.'
__all__ = [   # noqa: F405
    'BaseClient',
    'IAMClient',
    'create_client_from_env',
    'Client',
    'BasicAuthentication',
    'SoftLayerError',
    'SoftLayerAPIError',
    'SoftLayerListResult',
    'API_PUBLIC_ENDPOINT',
    'API_PRIVATE_ENDPOINT',
]



"""Set the LUN ID on an iSCSI volume."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('volume-id')
@click.argument('lun-id')
@environment.pass_env
def cli(env, volume_id, lun_id):
    """Set the LUN ID on an existing block storage volume.

    The LUN ID only takes effect during the Host Authorization process. It is
    recommended (but not necessary) to de-authorize all hosts before using this
    method. See `block access-revoke`.

    VOLUME_ID - the volume ID on which to set the LUN ID.

    LUN_ID - recommended range is an integer between 0 and 255. Advanced users
    can use an integer between 0 and 4095.
    """

    block_storage_manager = SoftLayer.BlockStorageManager(env.client)

    res = block_storage_manager.create_or_update_lun_id(volume_id, lun_id)

    if 'value' in res and lun_id == res['value']:
        click.echo(
            'Block volume with id %s is reporting LUN ID %s' % (res['volumeId'], res['value']))
    else:
        click.echo(
            'Failed to confirm the new LUN ID on volume %s' % (volume_id))



"""
    SoftLayer.CLI.template
    ~~~~~~~~~~~~~~~~~~~~~~
    Provides functions for loading/parsing and writing template files. Template
    files are used for storing CLI arguments in the form of a file to be used
    later with the --template option.

    :license: MIT, see LICENSE for more details.
"""

# pylint: disable=redefined-argument-from-local
import configparser
import io
import os.path


class TemplateCallback(object):
    """Callback to use to populate click arguments with a template."""

    def __init__(self, list_args=None):
        self.list_args = list_args or []

    def __call__(self, ctx, param, value):
        if value is None:
            return

        with open(os.path.expanduser(value), 'r', encoding="utf-8") as file_handle:
            config = configparser.ConfigParser()
            ini_str = '[settings]\n' + file_handle.read()
            ini_fp = io.StringIO(ini_str)
            config.read_file(ini_fp)

        # Merge template options with the options passed in
        args = {}
        for key, value in config.items('settings'):
            if key in self.list_args:
                value = value.split(',')

            if not args.get(key):
                args[key] = value

        if ctx.default_map is None:
            ctx.default_map = {}
        ctx.default_map.update(args)


def export_to_template(filename, args, exclude=None):
    """Exports given options to the given filename in INI format.

    :param filename: Filename to save options to
    :param dict args: Arguments to export
    :param list exclude (optional): Exclusion list for options that should not
                                    be exported
    """
    exclude = exclude or []
    exclude.append('config')
    exclude.append('really')
    exclude.append('format')
    exclude.append('debug')

    with open(filename, "w", encoding="utf-8") as template_file:
        for k, val in args.items():
            if val and k not in exclude:
                if isinstance(val, tuple):
                    val = ','.join(val)
                if isinstance(val, list):
                    val = ','.join(val)
                template_file.write('%s=%s\n' % (k, val))



"""Show current CLI configuration."""
# :license: MIT, see LICENSE for more details.

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import config
from SoftLayer.CLI import environment


@click.command(cls=SLCommand)
@environment.pass_env
def cli(env):
    """Show current configuration."""

    settings = config.get_settings_from_client(client=env.client, theme=env.theme)
    env.fout(config.config_table(settings))



listOriginPath = [
    {
        "header": "test.example.com",
        "httpPort": 80,
        "mappingUniqueId": "993419389425697",
        "origin": "10.10.10.1",
        "originType": "HOST_SERVER",
        "path": "/example",
        "status": "RUNNING"
    },
    {
        "header": "test.example.com",
        "httpPort": 80,
        "mappingUniqueId": "993419389425697",
        "origin": "10.10.10.1",
        "originType": "HOST_SERVER",
        "path": "/example1",
        "status": "RUNNING"
    }
]

createOriginPath = [
    {
        "httpPort": 80,
        "httpsPort": 81,
        "mappingUniqueId": "993419389425697",
        "origin": "10.10.10.1",
        "originType": "HOST_SERVER",
        "header": "test.example.com",
        "path": "/example",
        "status": "RUNNING",
        "bucketName": "test-bucket",
        'fileExtension': 'jpg',
        "performanceConfiguration": "Dynamic content acceleration",
        "dynamicContentAcceleration": {
                "detectionPath": "/abc.html",
                "prefetchEnabled": True,
                "mobileImageCompressionEnabled": True
            },
        "cacheKeyQueryRule": "include-all"
    }
]

deleteOriginPath = "Origin with path /example/videos/* has been deleted"



"""
    SoftLayer.firewall
    ~~~~~~~~~~~~~~~~~~
    Firewall Manager/helpers

    :license: MIT, see LICENSE for more details.
"""
from SoftLayer import exceptions
from SoftLayer import utils

RULE_MASK = ('mask[orderValue,action,destinationIpAddress,'
             'destinationIpSubnetMask,protocol,destinationPortRangeStart,'
             'destinationPortRangeEnd,sourceIpAddress,sourceIpSubnetMask,'
             'version,notes]')


def has_firewall(vlan):
    """Helper to determine whether or not a VLAN has a firewall.

    :param dict vlan: A dictionary representing a VLAN
    :returns: True if the VLAN has a firewall, false if it doesn't.
    """
    return bool(
        vlan.get('dedicatedFirewallFlag', None) or
        vlan.get('highAvailabilityFirewallFlag', None) or
        vlan.get('firewallInterfaces', None) or
        vlan.get('firewallNetworkComponents', None) or
        vlan.get('firewallGuestNetworkComponents', None)
    )


class FirewallManager(utils.IdentifierMixin, object):
    """Manages SoftLayer firewalls

    See product information here: https://www.ibm.com/cloud/network-security

    :param SoftLayer.API.BaseClient client: the client instance

    """

    def __init__(self, client):
        self.client = client
        self.account = self.client['Account']
        self.prod_pkg = self.client['Product_Package']

    def get_standard_package(self, server_id, is_virt=True):
        """Retrieves the standard firewall package for the virtual server.

        :param int server_id: The ID of the server to create the firewall for
        :param bool is_virt: True if the ID provided is for a virtual server,
                             False for a server
        :returns: A dictionary containing the standard virtual server firewall
                  package
        """

        firewall_port_speed = self._get_fwl_port_speed(server_id, is_virt)

        _value = "%s%s" % (firewall_port_speed, "Mbps Hardware Firewall")
        _filter = {'items': {'description': utils.query_filter(_value)}}

        return self.prod_pkg.getItems(id=0, filter=_filter)

    def get_dedicated_package(self, ha_enabled=False):
        """Retrieves the dedicated firewall package.

        :param bool ha_enabled: True if HA is to be enabled on the firewall
                                False for No HA
        :returns: A dictionary containing the dedicated virtual server firewall
                  package
        """

        fwl_filter = 'Hardware Firewall (Dedicated)'
        ha_fwl_filter = 'Hardware Firewall (High Availability)'
        _filter = utils.NestedDict({})
        if ha_enabled:
            _filter['items']['description'] = utils.query_filter(ha_fwl_filter)
        else:
            _filter['items']['description'] = utils.query_filter(fwl_filter)

        return self.prod_pkg.getItems(id=0, filter=_filter.to_dict())

    def cancel_firewall(self, firewall_id, dedicated=False):
        """Cancels the specified firewall.

        :param int firewall_id: Firewall ID to be cancelled.
        :param bool dedicated: If true, the firewall instance is dedicated,
                               otherwise, the firewall instance is shared.
        """

        fwl_billing = self._get_fwl_billing_item(firewall_id, dedicated)
        billing_item_service = self.client['Billing_Item']
        return billing_item_service.cancelService(id=fwl_billing['id'])

    def add_standard_firewall(self, server_id, is_virt=True):
        """Creates a firewall for the specified virtual/hardware server.

        :param int server_id: The ID of the server to create the firewall for
        :param bool is_virt: If true, will create the firewall for a virtual
                             server, otherwise for a hardware server.
        :returns: A dictionary containing the standard virtual server firewall
                  order
        """

        package = self.get_standard_package(server_id, is_virt)
        if is_virt:
            product_order = {
                'complexType': 'SoftLayer_Container_Product_Order_Network_'
                               'Protection_Firewall',
                'quantity': 1,
                'packageId': 0,
                'virtualGuests': [{'id': server_id}],
                'prices': [{'id': package[0]['prices'][0]['id']}]
            }
        else:
            product_order = {
                'complexType': 'SoftLayer_Container_Product_Order_Network_'
                               'Protection_Firewall',
                'quantity': 1,
                'packageId': 0,
                'hardware': [{'id': server_id}],
                'prices': [{'id': package[0]['prices'][0]['id']}]
            }
        return self.client['Product_Order'].placeOrder(product_order)

    def add_vlan_firewall(self, vlan_id, ha_enabled=False):
        """Creates a firewall for the specified vlan.

        :param int vlan_id: The ID of the vlan to create the firewall for
        :param bool ha_enabled: If True, an HA firewall will be created

        :returns: A dictionary containing the VLAN firewall order
        """

        package = self.get_dedicated_package(ha_enabled)
        product_order = {
            'complexType': 'SoftLayer_Container_Product_Order_Network_'
                           'Protection_Firewall_Dedicated',
            'quantity': 1,
            'packageId': 0,
            'vlanId': vlan_id,
            'prices': [{'id': package[0]['prices'][0]['id']}]
        }
        return self.client['Product_Order'].placeOrder(product_order)

    def _get_fwl_billing_item(self, firewall_id, dedicated=False):
        """Retrieves the billing item of the firewall.

        :param int firewall_id: Firewall ID to get the billing item for
        :param bool dedicated: whether the firewall is dedicated or standard
        :returns: A dictionary of the firewall billing item.
        """

        mask = 'mask[id,billingItem[id]]'
        if dedicated:
            firewall_service = self.client['Network_Vlan_Firewall']
        else:
            firewall_service = self.client['Network_Component_Firewall']
        firewall = firewall_service.getObject(id=firewall_id, mask=mask)
        if firewall is None:
            raise exceptions.SoftLayerError(f"Unable to find firewall {firewall_id}")
        if firewall.get('billingItem') is None:
            raise exceptions.SoftLayerError(f"Unable to find billing item for firewall {firewall_id}")

        return firewall['billingItem']

    def _get_fwl_port_speed(self, server_id, is_virt=True):
        """Determines the appropriate speed for a firewall.

        :param int server_id: The ID of server the firewall is for
        :param bool is_virt: True if the server_id is for a virtual server
        :returns: a integer representing the Mbps speed of a firewall
        """

        fwl_port_speed = 0
        if is_virt:
            mask = 'primaryNetworkComponent[maxSpeed]'
            svc = self.client['Virtual_Guest']
            primary = svc.getObject(mask=mask, id=server_id)
            fwl_port_speed = primary['primaryNetworkComponent']['maxSpeed']
        else:
            mask = 'id,maxSpeed,networkComponentGroup.networkComponents'
            svc = self.client['Hardware_Server']
            network_components = svc.getFrontendNetworkComponents(
                mask=mask, id=server_id)
            grouped = [interface['networkComponentGroup']['networkComponents']
                       for interface in network_components
                       if 'networkComponentGroup' in interface]
            ungrouped = [interface
                         for interface in network_components
                         if 'networkComponentGroup' not in interface]

            # For each group, sum the maxSpeeds of each compoment in the
            # group. Put the sum for each in a new list
            group_speeds = []
            for group in grouped:
                group_speed = 0
                for interface in group:
                    group_speed += interface['maxSpeed']
                group_speeds.append(group_speed)

            # The max speed of all groups is the max of the list
            max_grouped_speed = max(group_speeds)

            max_ungrouped = 0
            for interface in ungrouped:
                max_ungrouped = max(max_ungrouped, interface['maxSpeed'])

            fwl_port_speed = max(max_grouped_speed, max_ungrouped)

        return fwl_port_speed

    def get_firewalls(self):
        """Returns a list of all firewalls on the account.

        :returns: A list of firewalls on the current account.
        """

        mask = ('firewallNetworkComponents,'
                'networkVlanFirewall,'
                'dedicatedFirewallFlag,'
                'firewallGuestNetworkComponents,'
                'firewallInterfaces,'
                'firewallRules,'
                'highAvailabilityFirewallFlag')

        return [firewall
                for firewall in self.account.getNetworkVlans(mask=mask)
                if has_firewall(firewall)]

    def get_standard_fwl_rules(self, firewall_id):
        """Get the rules of a standard firewall.

        :param integer firewall_id: the instance ID of the standard firewall
        :returns: A list of the rules.
        """

        svc = self.client['Network_Component_Firewall']
        return svc.getRules(id=firewall_id, mask=RULE_MASK)

    def get_dedicated_fwl_rules(self, firewall_id):
        """Get the rules of a dedicated firewall.

        :param integer firewall_id: the instance ID of the dedicated firewall
        :returns: A list of the rules.
        """

        svc = self.client['Network_Vlan_Firewall']
        return svc.getRules(id=firewall_id, mask=RULE_MASK)

    def edit_dedicated_fwl_rules(self, firewall_id, rules):
        """Edit the rules for dedicated firewall.

        :param integer firewall_id: the instance ID of the dedicated firewall
        :param list rules: the rules to be pushed on the firewall as defined by
                           SoftLayer_Network_Firewall_Update_Request_Rule
        """

        mask = ('mask[networkVlan[firewallInterfaces'
                '[firewallContextAccessControlLists]]]')
        svc = self.client['Network_Vlan_Firewall']
        fwl = svc.getObject(id=firewall_id, mask=mask)
        network_vlan = fwl['networkVlan']

        for fwl1 in network_vlan['firewallInterfaces']:
            if fwl1['name'] == 'inside':
                continue

            for control_list in fwl1['firewallContextAccessControlLists']:
                if control_list['direction'] == 'out':
                    continue
                fwl_ctx_acl_id = control_list['id']

        template = {'firewallContextAccessControlListId': fwl_ctx_acl_id,
                    'rules': rules}

        svc = self.client['Network_Firewall_Update_Request']
        return svc.createObject(template)

    def edit_standard_fwl_rules(self, firewall_id, rules):
        """Edit the rules for standard firewall.

        :param integer firewall_id: the instance ID of the standard firewall
        :param list rules: the rules to be pushed on the firewall as defined by
                           SoftLayer_Network_Firewall_Update_Request_Rule
        """

        rule_svc = self.client['Network_Firewall_Update_Request']
        template = {'networkComponentFirewallId': firewall_id, 'rules': rules}

        return rule_svc.createObject(template)

    def get_instance(self, firewall_id, mask=None):
        """Get the firewall information

        :param integer firewall_id: the instance ID of the standard firewall
        """
        if not mask:
            mask = 'mask[firewallType,datacenter,managementCredentials,networkVlan,' \
                   'metricTrackingObject[data,type],networkGateway[insideVlans,members,privateIpAddress,' \
                   'publicIpAddress,publicIpv6Address,privateVlan,publicVlan,status]]'

        svc = self.client['Network_Vlan_Firewall']

        return svc.getObject(id=firewall_id, mask=mask)

    def get_firewalls_gatewalls(self):
        """Returns a list of all gateway firewalls (gatewalls) on the account.

        returns: A list of gateway firewalls (gatewalls) on the current account.
        """
        mask = 'mask[id,networkSpace,name,' \
               'networkFirewall[id,firewallType,datacenter[name]],' \
               'status[keyName],' \
               'insideVlans[id],' \
               'privateIpAddress[ipAddress],' \
               'publicVlan[id,primaryRouter[hostname]],' \
               'publicIpAddress[ipAddress],members[id,hardware[hostname]]]'
        _filter = {"networkGateways": {"networkFirewall": {"operation": "not null"}}}

        return self.account.getNetworkGateways(mask=mask, filter=_filter)

    def get_summary(self, identifier, start_date, end_date):
        """Returns the metric data for the date range provided

        :param integer metric_tracking_id
        """
        body = [{
            "keyName": "PUBLICIN",
            "summaryType": "sum"

        }, {
            "keyName": "PUBLICOUT",
            "summaryType": "sum"
        }]

        return self.client['Metric_Tracking_Object'].getSummaryData(start_date,
                                                                    end_date, body, 86400, id=identifier)



getAvailableRouters = [{
    "accountId": 1,
    "fullyQualifiedDomainName": "bcr01.dal01.softlayer.com",
    "hostname": "bcr01.dal01",
    "id": 1,
    "topLevelLocation": {
        "id": 3,
        "longName": "Dallas 1",
        "name": "dal01",
    }
}]

createObject = {
    "accountId": 123,
    "backendRouterId": 444,
    "createDate": "2019-01-18T16:08:44-06:00",
    "id": 5555,
    "modifyDate": None,
    "name": "test01",
    "ruleId": 1
}
getObject = {
    "createDate": "2019-01-17T14:36:42-06:00",
    "id": 1234,
    "name": "test-group",
    "backendRouter": {
        "hostname": "bcr01a.mex01",
        "id": 329266
    },
    "guests": [{
        "accountId": 123456789,
        "createDate": "2019-01-17T16:44:46-06:00",
        "domain": "test.com",
        "fullyQualifiedDomainName": "issues10691547765077.test.com",
        "hostname": "issues10691547765077",
        "id": 69131875,
        "maxCpu": 1,
        "maxMemory": 1024,
        "placementGroupId": 1234,
        "provisionDate": "2019-01-17T16:47:17-06:00",
        "activeTransaction": {
            "id": 107585077,
            "transactionStatus": {
                "friendlyName": "TESTING TXN",
                "name": "RECLAIM_WAIT"
            }
        },
        "globalIdentifier": "c786ac04-b612-4649-9d19-9662434eeaea",
        "primaryBackendIpAddress": "10.131.11.14",
        "primaryIpAddress": "169.57.70.180",
        "status": {
            "keyName": "DISCONNECTED",
            "name": "Disconnected"
        }
    }],
    "rule": {
        "id": 1,
        "keyName": "SPREAD",
        "name": "SPREAD"
    }
}

deleteObject = True



"""
    SoftLayer.CLI
    ~~~~~~~~~~~~~~
    Contains all code related to the CLI interface

    :license: MIT, see LICENSE for more details.
"""
# pylint: disable=w0401, invalid-name

from SoftLayer.CLI.helpers import *  # NOQA



"""Permanently removes an SSH key."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Permanently removes an SSH key."""
    mgr = SoftLayer.SshKeyManager(env.client)

    key_id = helpers.resolve_id(mgr.resolve_ids, identifier, 'SshKey')
    if not (env.skip_confirmations or formatting.no_going_back(key_id)):
        raise exceptions.CLIAbort('Aborted')

    mgr.delete_key(key_id)



"""Show all licenses."""
# :license: MIT, see LICENSE for more details.
import click

from SoftLayer import utils

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers import account


@click.command(cls=SLCommand)
@environment.pass_env
def cli(env):
    """Show all licenses."""

    manager = account.AccountManager(env.client)

    control_panel = manager.get_active_virtual_licenses()
    vmwares = manager.get_active_account_licenses()

    table_panel = formatting.KeyValueTable(['id', 'ip_address', 'manufacturer', 'software',
                                            'key', 'subnet', 'subnet notes'], title="Control Panel Licenses")

    table_vmware = formatting.KeyValueTable(['name', 'license_key', 'cpus', 'description',
                                             'manufacturer', 'requiredUser'], title="VMware Licenses")
    for panel in control_panel:
        table_panel.add_row([panel.get('id'), panel.get('ipAddress'),
                             utils.lookup(panel, 'softwareDescription', 'manufacturer'),
                             utils.trim_to(utils.lookup(panel, 'softwareDescription', 'longDescription'), 40),
                             panel.get('key'), utils.lookup(panel, 'subnet', 'broadcastAddress'),
                             utils.lookup(panel, 'subnet', 'note')])

    env.fout(table_panel)
    for vmware in vmwares:
        table_vmware.add_row([utils.lookup(vmware, 'softwareDescription', 'name'),
                              vmware.get('key'), vmware.get('capacity'),
                              utils.lookup(vmware, 'billingItem', 'description'),
                              utils.lookup(vmware, 'softwareDescription', 'manufacturer'),
                              utils.lookup(vmware, 'softwareDescription', 'requiredUser')])

    env.fout(table_vmware)



"""List virtual server credentials."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """List virtual server credentials."""

    vsi = SoftLayer.VSManager(env.client)
    vs_id = helpers.resolve_id(vsi.resolve_ids, identifier, 'VS')
    instance = vsi.get_instance(vs_id)

    table = formatting.Table(['username', 'password', 'Software', 'Version'])

    for item in instance['operatingSystem']['passwords']:
        table.add_row([item['username'], item['password'],
                       instance['operatingSystem']['softwareLicense']['softwareDescription']['referenceCode'],
                       instance['operatingSystem']['softwareLicense']['softwareDescription']['version']])
    env.fout(table)



"""Create a placement group"""

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import helpers
from SoftLayer.managers.vs_placement import PlacementManager as PlacementManager


@click.command(cls=SLCommand)
@click.option('--name', type=click.STRING, required=True, prompt=True, help="Name for this new placement group.")
@click.option('--backend_router', '-b', required=True, prompt=True,
              help="backendRouter, can be either the hostname or id.")
@click.option('--rule', '-r', required=True, prompt=True,
              help="The keyName or Id of the rule to govern this placement group.")
@environment.pass_env
def cli(env, **args):
    """Create a placement group."""
    manager = PlacementManager(env.client)
    backend_router_id = helpers.resolve_id(manager.get_backend_router_id_from_hostname,
                                           args.get('backend_router'),
                                           'backendRouter')
    rule_id = helpers.resolve_id(manager.get_rule_id_from_name, args.get('rule'), 'Rule')
    placement_object = {
        'name': args.get('name'),
        'backendRouterId': backend_router_id,
        'ruleId': rule_id
    }

    result = manager.create(placement_object)
    click.secho(f"Successfully created placement group: ID: {result['id']}, Name: {result['name']}", fg='green')



"""CLI configuration."""
# :license: MIT, see LICENSE for more details.

from SoftLayer.CLI import formatting


def _resolve_transport(transport):
    """recursively look for transports which refer to other transports."""
    nested_transport = getattr(transport, 'transport', None)
    if nested_transport is not None:
        return nested_transport

    return _resolve_transport(nested_transport)


def get_settings_from_client(client, theme):
    """Pull out settings from a SoftLayer.BaseClient instance.

    :param client: SoftLayer.BaseClient instance
    """
    settings = {
        'username': '',
        'api_key': '',
        'timeout': '',
        'endpoint_url': '',
        'theme': theme
    }
    try:
        settings['username'] = client.auth.username
        settings['api_key'] = client.auth.api_key
    except AttributeError:
        pass

    transport = _resolve_transport(client.transport)
    try:
        settings['timeout'] = transport.timeout
        settings['endpoint_url'] = transport.endpoint_url
    except AttributeError:
        pass

    return settings


def config_table(settings):
    """Returns a config table."""
    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'
    table.add_row(['Username', settings['username'] or 'not set'])
    table.add_row(['API Key', settings['api_key'] or 'not set'])
    table.add_row(['Endpoint URL', settings['endpoint_url'] or 'not set'])
    table.add_row(['Timeout', settings['timeout'] or 'not set'])
    table.add_row(['Theme', settings['theme'] or 'not set'])
    return table



"""Show all Provisioning Scripts."""
# :license: MIT, see LICENSE for more details.
import click


from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers import account


@click.command(cls=SLCommand)
@environment.pass_env
def cli(env):
    """Show all Provisioning Scripts."""

    manager = account.AccountManager(env.client)
    hooks = manager.get_provisioning_scripts()

    table = formatting.Table(["Id", "Name", "Uri"])

    for hook in hooks:
        table.add_row([hook.get('id'), hook.get('name'), hook.get('uri')])

    env.fout(table)



"""Cancel a subnet."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Cancel a subnet."""

    mgr = SoftLayer.NetworkManager(env.client)
    subnet_id = helpers.resolve_id(mgr.resolve_subnet_ids, identifier,
                                   name='subnet')

    if not (env.skip_confirmations or formatting.no_going_back(subnet_id)):
        raise exceptions.CLIAbort('Aborted')

    mgr.cancel_subnet(subnet_id)



"""Refresh a duplicate volume with a snapshot from its parent."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import environment


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('volume_id')
@click.argument('snapshot_id')
@click.option('--force-refresh', '-f', is_flag=True, default=False, show_default=True,
              help="Cancel current refresh process and initiates the new refresh.")
@environment.pass_env
def cli(env, volume_id, snapshot_id, force_refresh):
    """Refresh a duplicate volume with a snapshot from its parent.

    EXAMPLE::

            slcli block volume-refresh VOLUME_ID SNAPSHOT_ID
            Refresh a duplicate VOLUME_ID with a snapshot from its parent SNAPSHOT_ID.
    """
    block_manager = SoftLayer.BlockStorageManager(env.client)
    resp = block_manager.refresh_dupe(volume_id, snapshot_id, force_refresh)

    click.echo(resp)



"""Compute images."""
# :license: MIT, see LICENSE for more details.
from SoftLayer.CLI import formatting


MASK = ('id,createDate,note,accountId,name,globalIdentifier,parentId,publicFlag,flexImageFlag,'
        'imageType,children[blockDevices[diskImage[softwareReferences[softwareDescription]]]]')
DETAIL_MASK = ('id,createDate,note,accountId,name,globalIdentifier,parentId,publicFlag,flexImageFlag,'
               'imageType,firstChild,children[id,blockDevicesDiskSpaceTotal,datacenter,'
               'transaction[transactionGroup,transactionStatus],'
               'blockDevices[diskImage[capacity,name,units,softwareReferences[softwareDescription]],diskSpace]],'
               'status,transaction,accountReferences')
PUBLIC_TYPE = formatting.FormattedItem('PUBLIC', 'Public')
PRIVATE_TYPE = formatting.FormattedItem('PRIVATE', 'Private')



# https://sldn.softlayer.com/reference/datatypes/SoftLayer_Network_LBaaS_Listener/
updateLoadBalancerProtocols = {'accountId': 1234,
                               'address': '01-307608-ams01.clb.appdomain.cloud',
                               'createDate': '2019-08-12T07:49:43-06:00',
                               'id': 1111111,
                               'isPublic': 0,
                               'locationId': 265592,
                               'modifyDate': '2019-08-13T16:26:06-06:00',
                               'name': 'dcabero-01',
                               'operatingStatus': 'ONLINE',
                               'provisioningStatus': 'ACTIVE',
                               'type': 0,
                               'useSystemPublicIpPool': 1,
                               'uuid': '1a1aa111-4474-4e16-9f02-4de959229b85',
                               'listenerCount': 4,
                               'memberCount': 1,
                               'datacenter': {
                                   'id': 265592,
                                   'longName': 'Amsterdam 1',
                                   'name': 'ams01',
                                   'statusId': 2
                               }}
deleteLoadBalancerProtocols = {'accountId': 1234,
                               'address': '01-307608-ams01.clb.appdomain.cloud',
                               'createDate': '2019-08-12T07:49:43-06:00',
                               'id': 1111111,
                               'isPublic': 0,
                               'locationId': 265592,
                               'modifyDate': '2019-08-13T16:26:06-06:00',
                               'name': 'dcabero-01',
                               'operatingStatus': 'ONLINE',
                               'provisioningStatus': 'ACTIVE',
                               'type': 0,
                               'useSystemPublicIpPool': 1,
                               'uuid': '1a1aa111-4474-4e16-9f02-4de959229b85',
                               'listenerCount': 4,
                               'memberCount': 1,
                               'datacenter': {
                                   'id': 265592,
                                   'longName': 'Amsterdam 1',
                                   'name': 'ams01',
                                   'statusId': 2
                               }}

getL7Policies = [
    {'action': 'REJECT',
     'createDate': '2021-09-08T15:08:35-06:00',
     'id': 123456,
     'modifyDate': None,
     'name': 'test-reject',
     'priority': 2,
     'redirectL7PoolId': None,
     'uuid': '123mock-1234-43c9-b659-12345678mock'
     },
    {'action': 'REDIRECT_HTTPS',
     'createDate': '2021-09-08T15:03:53-06:00',
     'id': 432922,
     'modifyDate': None,
     'name': 'test-policy-https-1',
     'priority': 0,
     'redirectL7PoolId': None,
     'redirectUrl': 'url-test-uuid-mock-1234565',
     'uuid': 'test-uuid-mock-1234565'
     }
]



"""List guests which are in a dedicated host server."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import columns as column_helper
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers
from SoftLayer import utils

COLUMNS = [
    column_helper.Column('guid', ('globalIdentifier',)),
    column_helper.Column('cpu', ('maxCpu',)),
    column_helper.Column('memory', ('maxMemory',)),
    column_helper.Column('datacenter', ('datacenter', 'name')),
    column_helper.Column('primary_ip', ('primaryIpAddress',)),
    column_helper.Column('backend_ip', ('primaryBackendIpAddress',)),
    column_helper.Column(
        'created_by',
        lambda created_by: utils.lookup(created_by, 'billingItem', 'orderItem', 'order', 'userRecord', 'username'),
        mask='billingItem[id,orderItem[id,order[id,userRecord[username]]]]'),
    column_helper.Column('power_state', ('powerState', 'name')),
    column_helper.Column(
        'tags',
        lambda server: formatting.tags(server.get('tagReferences')),
        mask="tagReferences.tag.name"),
]

DEFAULT_COLUMNS = [
    'id',
    'hostname',
    'domain',
    'primary_ip',
    'backend_ip',
    'power_state'
]


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--cpu', '-c', help='Number of CPU cores', type=click.INT)
@click.option('--domain', '-D', help='Domain portion of the FQDN')
@click.option('--hostname', '-H', help='Host portion of the FQDN')
@click.option('--memory', '-m', help='Memory in mebibytes', type=click.INT)
@helpers.multi_option('--tag', help='Filter by tags')
@click.option('--sortby',
              help='Column to sort by',
              default='hostname',
              show_default=True)
@click.option('--columns',
              callback=column_helper.get_formatter(COLUMNS),
              help='Columns to display. [options: %s]'
              % ', '.join(column.name for column in COLUMNS),
              default=','.join(DEFAULT_COLUMNS),
              show_default=True)
@environment.pass_env
def cli(env, identifier, sortby, cpu, domain, hostname, memory, tag, columns):
    """List guests which are in a dedicated host server."""

    mgr = SoftLayer.DedicatedHostManager(env.client)
    guests = mgr.list_guests(host_id=identifier,
                             cpus=cpu,
                             hostname=hostname,
                             domain=domain,
                             memory=memory,
                             tags=tag,
                             mask=columns.mask())

    table = formatting.Table(columns.columns)
    table.sortby = sortby

    for guest in guests:
        table.add_row([value or formatting.blank()
                       for value in columns.row(guest)])

    env.fout(table)



"""Set the user VPN password."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import helpers


@click.command(cls=SLCommand, )
@click.argument('identifier')
@click.option('--password', required=True, help="Your new VPN password")
@environment.pass_env
def cli(env, identifier, password):
    """Set the user VPN password.

    Example: slcli user vpn-password 123456 --password=Mypassword1.
    """

    mgr = SoftLayer.UserManager(env.client)
    user_id = helpers.resolve_id(mgr.resolve_ids, identifier, 'username')
    result = mgr.update_vpn_password(user_id, password)
    if result:
        click.secho("Successfully updated user VPN password.", fg='green')



"""Report on Resources in closing datacenters"""
import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer import utils


@click.command(cls=SLCommand, short_help="""Report on Resources in closing datacenters""")
@environment.pass_env
def cli(env):
    """Report on Resources in closing datacenters

    Displays a list of Datacenters soon to be shutdown, and any resources on the account
    in those locations
    """

    closing_filter = {
        'capabilities': {
            'operation': 'in',
            'options': [{'name': 'data', 'value': ['CLOSURE_ANNOUNCED']}]
        },
        'name': {
            'operation': 'orderBy',
            'options': [{'name': 'sort', 'value': ['DESC']}]
        }
    }
    mask = """mask[name, datacenterLongName, frontendRouterId, capabilities, datacenterId, backendRouterId,
backendRouterName, frontendRouterName]"""
    closing_pods = env.client.call('SoftLayer_Network_Pod', 'getAllObjects', mask=mask, filter=closing_filter)
    # Find all VLANs in the POD that is going to close.
    search = "_objectType:SoftLayer_Network_Vlan  primaryRouter.hostname: \"{}\" || primaryRouter.hostname: \"{}\""
    resource_mask = """mask[
        resource(SoftLayer_Network_Vlan)[
            id,fullyQualifiedName,name,note,vlanNumber,networkSpace,
            virtualGuests[id,fullyQualifiedDomainName,billingItem[cancellationDate]],
            hardware[id,fullyQualifiedDomainName,billingItem[cancellationDate]],
            networkVlanFirewall[id,primaryIpAddress,billingItem[cancellationDate]],
            privateNetworkGateways[id,name,networkSpace],
            publicNetworkGateways[id,name,networkSpace]
        ]
    ]
    """
    table_title = "Resources in closing datacenters"
    resource_table = formatting.Table(["Id", "Name", "Public VLAN", "Private VLAN", "Type", "Datacenter",
                                       "POD", "Cancellation Date"], title=table_title)
    resource_table.align = 'l'
    for pod in closing_pods:
        resources = {'hardware': {}, 'virtual': {}, 'firewall': {}, 'gateway': {}}
        vlans = env.client.call('SoftLayer_Search', 'advancedSearch',
                                search.format(pod.get('backendRouterName'), pod.get('frontendRouterName')),
                                iter=True, mask=resource_mask)
        # Go through the vlans and coalate the resources into a data structure that is easy to print out
        for vlan in vlans:
            resources = process_vlan(vlan.get('resource', {}), resources)

        # Go through each resource and add it to the table
        for resource_type, resource_values in resources.items():
            for resource_id, resource_object in resource_values.items():
                resource_table.add_row([
                    resource_id,
                    resource_object['name'],
                    resource_object['vlan'].get('PUBLIC', '-'),
                    resource_object['vlan'].get('PRIVATE', '-'),
                    resource_type,
                    pod.get('datacenterLongName'),
                    pod.get('backendRouterName'),
                    resource_object['cancelDate']
                ])

    env.fout(resource_table)


# returns a Table Row for a given resource
def process_vlan(vlan, resources=None):
    """Takes in a vlan object and pulls out the needed resources"""
    if resources is None:
        resources = {'hardware': {}, 'virtual': {}, 'firewall': {}, 'gateway': {}}

    type_x = "virtual"
    for obj_x in vlan.get('virtualGuests', {}):
        existing = resources[type_x].get(obj_x.get('id'))
        resources[type_x][obj_x['id']] = build_resource_object('fullyQualifiedDomainName', vlan, obj_x, existing)

    type_x = 'hardware'
    for obj_x in vlan.get('hardware', {}):
        existing = resources[type_x].get(obj_x.get('id'))
        resources[type_x][obj_x['id']] = build_resource_object('fullyQualifiedDomainName', vlan, obj_x, existing)

    type_x = 'firewall'
    for obj_x in vlan.get('networkVlanFirewall', {}):
        existing = resources[type_x].get(obj_x.get('id'))
        resources[type_x][obj_x['id']] = build_resource_object('primaryIpAddress', vlan, obj_x, existing)

    type_x = 'gateway'
    for obj_x in vlan.get('privateNetworkGateways', {}):
        existing = resources[type_x].get(obj_x.get('id'))
        resources[type_x][obj_x['id']] = build_resource_object('name', vlan, obj_x, existing)
    for obj_x in vlan.get('publicNetworkGateways', {}):
        existing = resources[type_x].get(obj_x.get('id'))
        resources[type_x][obj_x['id']] = build_resource_object('name', vlan, obj_x, existing)

    return resources


def build_resource_object(name_property, vlan, resource, entry):
    """builds out a resource object and puts the required values in the right place.

    :param: name_property is what property to use as the name from resource
    :param: vlan is the vlan object
    :param: resource has the data we want
    :param: entry is for any existing data
    """
    new_entry = {
        'id': resource.get('id'),
        'name': resource.get(name_property),
        'vlan': {vlan.get('networkSpace'): vlan.get('vlanNumber')},
        'cancelDate': utils.clean_time(utils.lookup(resource, 'billingItem', 'cancellationDate'))
    }
    if entry:
        entry['vlan'][vlan.get('networkSpace')] = vlan.get('vlanNumber')
    else:
        entry = new_entry

    return entry



getObject = {
    'endDate': '2019-03-18T17:00:00-06:00',
    'id': 1234,
    'lastImpactedUserCount': 417756,
    'modifyDate': '2019-03-12T15:32:48-06:00',
    'recoveryTime': None,
    'startDate': '2019-03-18T16:00:00-06:00',
    'subject': 'Public Website Maintenance',
    'summary': 'Blah Blah Blah',
    'systemTicketId': 76057381,
    'acknowledgedFlag': False,
    'attachments': [],
    'impactedResources': [{
        'resourceType': 'Server',
        'resourceTableId': 12345,
        'hostname': 'test',
        'privateIp': '10.0.0.1',
        'filterLable': 'Server'
    }],
    'notificationOccurrenceEventType': {'keyName': 'PLANNED'},
    'statusCode': {'keyName': 'PUBLISHED', 'name': 'Published'},
    'updates': [{
        'contents': 'More Blah Blah',
        'createDate': '2019-03-12T13:07:22-06:00',
        'endDate': None, 'startDate': '2019-03-12T13:07:22-06:00'
    }]
}

getAllObjects = [getObject]

acknowledgeNotification = True



"""Get the snapshots space usage threshold warning flag setting for specific volume"""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import environment


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('volume_id')
@environment.pass_env
def cli(env, volume_id):
    """Get snapshots space usage threshold warning flag setting for a given volume."""
    block_manager = SoftLayer.BlockStorageManager(env.client)
    enabled = block_manager.get_volume_snapshot_notification_status(volume_id)

    if enabled == 0:
        click.echo(f"Disabled: Snapshots space usage threshold is disabled for volume {volume_id}")
    else:
        click.echo(f"Enabled: Snapshots space usage threshold is enabled for volume {volume_id}")



"""Set note for an existing block storage volume."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('volume-id')
@click.option('--note', '-n',
              type=str,
              required=True,
              help='Public notes related to a Storage volume')
@environment.pass_env
def cli(env, volume_id, note):
    """Set note for an existing block storage volume.

    EXAMPLE::

            slcli block volume-set-note 12345678 --note 'this is my note'
    """
    block_manager = SoftLayer.BlockStorageManager(env.client)
    block_volume_id = helpers.resolve_id(block_manager.resolve_ids, volume_id, 'Block Volume')

    result = block_manager.volume_set_note(block_volume_id, note)

    if result:
        click.echo("Set note successfully!")

    else:
        click.echo("Note could not be set! Please verify your options and try again.")



"""Edit details of a security group."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('group_id')
@click.option('--name', '-n',
              help="The name of the security group")
@click.option('--description', '-d',
              help="The description of the security group")
@environment.pass_env
def cli(env, group_id, name, description):
    """Edit details of a security group."""
    mgr = SoftLayer.NetworkManager(env.client)
    data = {}
    if name:
        data['name'] = name
    if description:
        data['description'] = description

    if not mgr.edit_securitygroup(group_id, **data):
        raise exceptions.CLIAbort("Failed to edit security group")



"""
    SoftLayer.transports.transport
    ~~~~~~~~~~~~~~~~~~~~
    Common functions for transporting API requests

    :license: MIT, see LICENSE for more details.
"""
import base64
import json
import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry


from SoftLayer import utils


def get_session(user_agent):
    """Sets up urllib sessions"""

    client = requests.Session()
    client.headers.update({
        'Content-Type': 'application/json',
        'User-Agent': user_agent,
    })
    retry = Retry(total=3, connect=1, backoff_factor=1)
    adapter = HTTPAdapter(max_retries=retry)
    client.mount('https://', adapter)
    return client


# transports.Request does have a lot of instance attributes. :(
# pylint: disable=too-many-instance-attributes
class Request(object):
    """Transport request object."""

    def __init__(self):
        #: API service name. E.G. SoftLayer_Account
        self.service = None

        #: API method name. E.G. getObject
        self.method = None

        #: API Parameters.
        self.args = tuple()

        #: URL Parameters, used for the REST Transport
        self.params = None

        #: API headers, used for authentication, masks, limits, offsets, etc.
        self.headers = {}

        #: Transport user.
        self.transport_user = None

        #: Transport password.
        self.transport_password = None

        #: Transport headers.
        self.transport_headers = {}

        #: False -> Don't verify the SSL certificate
        #: True -> Verify the SSL certificate
        #: Path String -> Verify the SSL certificate with the .pem file at path
        self.verify = None

        #: Client certificate file path. (Used by X509Authentication)
        self.cert = None

        #: InitParameter/identifier of an object.
        self.identifier = None

        #: SoftLayer mask (dict or string).
        self.mask = None

        #: SoftLayer Filter (dict).
        self.filter = None

        #: Integer result limit.
        self.limit = None

        #: Integer result offset.
        self.offset = None

        #: Integer call start time
        self.start_time = None

        #: Integer call end time
        self.end_time = None

        #: String full url
        self.url = None

        #: String result of api call
        self.result = None

        #: String payload to send in
        self.payload = None

        #: Exception any exceptions that got caught
        self.exception = None

    def __repr__(self):
        """Prints out what this call is all about"""
        pretty_mask = utils.clean_string(self.mask)
        pretty_filter = self.filter
        clean_args = self.args
        # Passwords can show up here, so censor them before logging.
        if self.method in ["performExternalAuthentication", "refreshEncryptedToken",
                           "getPortalLoginToken", "getEncryptedSessionToken"]:
            clean_args = "*************"
        param_string = (f"id={self.identifier}, mask='{pretty_mask}', filter='{pretty_filter}', args={clean_args}, "
                        f"limit={self.limit}, offset={self.offset}")
        return "{service}::{method}({params})".format(
            service=self.service, method=self.method, params=param_string)

    def special_rest_params(self):
        """This method is to handle the edge case of SoftLayer_User_Employee::getEncryptedSessionToken

        Added this method here since it was a little easier to change the data as needed this way.
        """
        if self.method == "getEncryptedSessionToken" and self.service == "SoftLayer_User_Employee":
            if len(self.args) < 3:
                return
            self.params = {"remoteToken": self.args[2]}
            self.transport_user = self.args[0]
            self.transport_password = self.args[1]
            self.args = []


class SoftLayerListResult(list):
    """A SoftLayer API list result."""

    def __init__(self, items=None, total_count=0):

        #: total count of items that exist on the server. This is useful when
        #: paginating through a large list of objects.
        self.total_count = total_count
        super().__init__(items)

    def get_total_items(self):
        """A simple getter to totalCount, but its called getTotalItems since that is the header returned"""
        return self.total_count


def _proxies_dict(proxy):
    """Makes a proxy dict appropriate to pass to requests."""
    if not proxy:
        return None
    return {'http': proxy, 'https': proxy}


def _format_object_mask(objectmask):
    """Format the new style object mask.

    This wraps the user mask with mask[USER_MASK] if it does not already
    have one. This makes it slightly easier for users.

    :param objectmask: a string-based object mask

    """
    objectmask = objectmask.strip()

    if (not objectmask.startswith('mask') and
            not objectmask.startswith('[') and
            not objectmask.startswith('filteredMask')):
        objectmask = "mask[%s]" % objectmask
    return objectmask


class ComplexEncoder(json.JSONEncoder):
    """ComplexEncoder helps jsonencoder deal with byte strings"""

    def default(self, o):
        """Encodes o as JSON"""

        # Base64 encode bytes type objects.
        if isinstance(o, bytes):
            base64_bytes = base64.b64encode(o)
            return base64_bytes.decode("utf-8")
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, o)
